# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
import time

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from taskflow import task
from taskflow.types import failure
import tenacity

from octavia.common import constants
from octavia.common import data_models
from octavia.common import utils
from octavia.controller.worker import task_utils
from octavia.db import api as db_apis
from octavia.db import repositories as repo
from octavia.network import base
from octavia.network import data_models as n_data_models

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class BaseNetworkTask(task.Task):
    """Base task to load drivers common to the tasks."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._network_driver = None
        self.task_utils = task_utils.TaskUtils()
        self.loadbalancer_repo = repo.LoadBalancerRepository()
        self.amphora_repo = repo.AmphoraRepository()

    @property
    def network_driver(self):
        if self._network_driver is None:
            self._network_driver = utils.get_network_driver()
        return self._network_driver


class CalculateAmphoraDelta(BaseNetworkTask):

    default_provides = constants.DELTA

    def execute(self, loadbalancer, amphora, availability_zone,
                vrrp_port=None):
        LOG.debug("Calculating network delta for amphora id: %s",
                  amphora.get(constants.ID))

        if vrrp_port is None:
            vrrp_port = self.network_driver.get_port(
                amphora[constants.VRRP_PORT_ID])
            vrrp_port_network_id = vrrp_port.network_id
        else:
            vrrp_port_network_id = vrrp_port[constants.NETWORK_ID]

        # Figure out what networks we want
        # seed with lb network(s)
        if (availability_zone and
                availability_zone.get(constants.MANAGEMENT_NETWORK)):
            management_nets = [
                availability_zone.get(constants.MANAGEMENT_NETWORK)]
        else:
            management_nets = CONF.controller_worker.amp_boot_network_list
        desired_network_ids = {vrrp_port_network_id}.union(management_nets)
        db_lb = self.loadbalancer_repo.get(
            db_apis.get_session(), id=loadbalancer[constants.LOADBALANCER_ID])
        for pool in db_lb.pools:
            member_networks = [
                self.network_driver.get_subnet(member.subnet_id).network_id
                for member in pool.members
                if member.subnet_id
            ]
            desired_network_ids.update(member_networks)

        nics = self.network_driver.get_plugged_networks(
            amphora[constants.COMPUTE_ID])
        # assume we don't have two nics in the same network
        actual_network_nics = dict((nic.network_id, nic) for nic in nics)

        del_ids = set(actual_network_nics) - desired_network_ids
        delete_nics = list(
            n_data_models.Interface(network_id=net_id) for net_id in del_ids)

        add_ids = desired_network_ids - set(actual_network_nics)
        add_nics = list(n_data_models.Interface(
            network_id=net_id) for net_id in add_ids)
        delta = n_data_models.Delta(
            amphora_id=amphora[constants.ID],
            compute_id=amphora[constants.COMPUTE_ID],
            add_nics=add_nics, delete_nics=delete_nics)
        return delta.to_dict(recurse=True)


class CalculateDelta(BaseNetworkTask):
    """Task to calculate the delta between

    the nics on the amphora and the ones
    we need. Returns a list for
    plumbing them.
    """

    default_provides = constants.DELTAS

    def execute(self, loadbalancer, availability_zone):
        """Compute which NICs need to be plugged

        for the amphora to become operational.

        :param loadbalancer: the loadbalancer to calculate deltas for all
                             amphorae
        :param availability_zone: availability zone metadata dict

        :returns: dict of octavia.network.data_models.Delta keyed off amphora
                  id
        """

        calculate_amp = CalculateAmphoraDelta()
        deltas = {}
        db_lb = self.loadbalancer_repo.get(
            db_apis.get_session(), id=loadbalancer[constants.LOADBALANCER_ID])
        for amphora in filter(
            lambda amp: amp.status == constants.AMPHORA_ALLOCATED,
                db_lb.amphorae):

            delta = calculate_amp.execute(loadbalancer, amphora.to_dict(),
                                          availability_zone)
            deltas[amphora.id] = delta
        return deltas


class GetPlumbedNetworks(BaseNetworkTask):
    """Task to figure out the NICS on an amphora.

    This will likely move into the amphora driver
    :returns: Array of networks
    """

    default_provides = constants.NICS

    def execute(self, amphora):
        """Get plumbed networks for the amphora."""

        LOG.debug("Getting plumbed networks for amphora id: %s",
                  amphora[constants.ID])

        return self.network_driver.get_plugged_networks(
            amphora[constants.COMPUTE_ID])


class PlugNetworks(BaseNetworkTask):
    """Task to plug the networks.

    This uses the delta to add all missing networks/nics
    """

    def execute(self, amphora, delta):
        """Update the amphora networks for the delta."""

        LOG.debug("Plug or unplug networks for amphora id: %s",
                  amphora[constants.ID])

        if not delta:
            LOG.debug("No network deltas for amphora id: %s",
                      amphora[constants.ID])
            return

        # add nics
        for nic in delta[constants.ADD_NICS]:
            self.network_driver.plug_network(amphora[constants.COMPUTE_ID],
                                             nic[constants.NETWORK_ID])

    def revert(self, amphora, delta, *args, **kwargs):
        """Handle a failed network plug by removing all nics added."""

        LOG.warning("Unable to plug networks for amp id %s",
                    amphora[constants.ID])
        if not delta:
            return

        for nic in delta[constants.ADD_NICS]:
            try:
                self.network_driver.unplug_network(
                    amphora[constants.COMPUTE_ID],
                    nic[constants.NETWORK_ID])
            except base.NetworkNotFound:
                pass


class UnPlugNetworks(BaseNetworkTask):
    """Task to unplug the networks

    Loop over all nics and unplug them
    based on delta
    """

    def execute(self, amphora, delta):
        """Unplug the networks."""

        LOG.debug("Unplug network for amphora")
        if not delta:
            LOG.debug("No network deltas for amphora id: %s",
                      amphora[constants.ID])
            return

        for nic in delta[constants.DELETE_NICS]:
            try:
                self.network_driver.unplug_network(
                    amphora[constants.COMPUTE_ID], nic[constants.NETWORK_ID])
            except base.NetworkNotFound:
                LOG.debug("Network %d not found", nic[constants.NETWORK_ID])
            except Exception:
                LOG.exception("Unable to unplug network")
                # TODO(xgerman) follow up if that makes sense


class GetMemberPorts(BaseNetworkTask):

    def execute(self, loadbalancer, amphora):
        vip_port = self.network_driver.get_port(loadbalancer['vip_port_id'])
        member_ports = []
        interfaces = self.network_driver.get_plugged_networks(
            amphora[constants.COMPUTE_ID])
        for interface in interfaces:
            port = self.network_driver.get_port(interface.port_id)
            if vip_port.network_id == port.network_id:
                continue
            port.network = self.network_driver.get_network(port.network_id)
            for fixed_ip in port.fixed_ips:
                if amphora['lb_network_ip'] == fixed_ip.ip_address:
                    break
                fixed_ip.subnet = self.network_driver.get_subnet(
                    fixed_ip.subnet_id)
            # Only add the port to the list if the IP wasn't the mgmt IP
            else:
                member_ports.append(port)
        return member_ports


class HandleNetworkDelta(BaseNetworkTask):
    """Task to plug and unplug networks

    Plug or unplug networks based on delta
    """

    def execute(self, amphora, delta):
        """Handle network plugging based off deltas."""
        added_ports = {}
        added_ports[amphora[constants.ID]] = []
        for nic in delta[constants.ADD_NICS]:
            interface = self.network_driver.plug_network(
                delta[constants.COMPUTE_ID], nic[constants.NETWORK_ID])
            port = self.network_driver.get_port(interface.port_id)
            port.network = self.network_driver.get_network(port.network_id)
            for fixed_ip in port.fixed_ips:
                fixed_ip.subnet = self.network_driver.get_subnet(
                    fixed_ip.subnet_id)
            added_ports[amphora[constants.ID]].append(port.to_dict(
                recurse=True))
        for nic in delta[constants.DELETE_NICS]:
            try:
                self.network_driver.unplug_network(
                    delta[constants.COMPUTE_ID], nic[constants.NETWORK_ID])
            except base.NetworkNotFound:
                LOG.debug("Network %d not found ", nic[constants.NETWORK_ID])
            except Exception:
                LOG.exception("Unable to unplug network")
        return added_ports

    def revert(self, result, amphora, delta, *args, **kwargs):
        """Handle a network plug or unplug failures."""

        if isinstance(result, failure.Failure):
            return

        if not delta:
            return

        LOG.warning("Unable to plug networks for amp id %s",
                    delta['amphora_id'])

        for nic in delta[constants.ADD_NICS]:
            try:
                self.network_driver.unplug_network(delta[constants.COMPUTE_ID],
                                                   nic[constants.NETWORK_ID])
            except Exception:
                pass


class HandleNetworkDeltas(BaseNetworkTask):
    """Task to plug and unplug networks

    Loop through the deltas and plug or unplug
    networks based on delta
    """

    def execute(self, deltas):
        """Handle network plugging based off deltas."""
        added_ports = {}
        for amp_id, delta in deltas.items():
            added_ports[amp_id] = []
            for nic in delta[constants.ADD_NICS]:
                interface = self.network_driver.plug_network(
                    delta[constants.COMPUTE_ID], nic[constants.NETWORK_ID])
                port = self.network_driver.get_port(interface.port_id)
                port.network = self.network_driver.get_network(port.network_id)
                for fixed_ip in port.fixed_ips:
                    fixed_ip.subnet = self.network_driver.get_subnet(
                        fixed_ip.subnet_id)
                added_ports[amp_id].append(port.to_dict(recurse=True))
            for nic in delta[constants.DELETE_NICS]:
                try:
                    self.network_driver.unplug_network(
                        delta[constants.COMPUTE_ID],
                        nic[constants.NETWORK_ID])
                except base.NetworkNotFound:
                    LOG.debug("Network %d not found ",
                              nic[constants.NETWORK_ID])
                except Exception:
                    LOG.exception("Unable to unplug network")
        return added_ports

    def revert(self, result, deltas, *args, **kwargs):
        """Handle a network plug or unplug failures."""

        if isinstance(result, failure.Failure):
            return
        for amp_id, delta in deltas.items():
            LOG.warning("Unable to plug networks for amp id %s",
                        delta[constants.AMPHORA_ID])
            if not delta:
                return

            for nic in delta[constants.ADD_NICS]:
                try:
                    self.network_driver.unplug_network(
                        delta[constants.COMPUTE_ID],
                        nic[constants.NETWORK_ID])
                except base.NetworkNotFound:
                    pass


class PlugVIP(BaseNetworkTask):
    """Task to plumb a VIP."""

    def execute(self, loadbalancer):
        """Plumb a vip to an amphora."""

        LOG.debug("Plumbing VIP for loadbalancer id: %s",
                  loadbalancer[constants.LOADBALANCER_ID])
        db_lb = self.loadbalancer_repo.get(
            db_apis.get_session(),
            id=loadbalancer[constants.LOADBALANCER_ID])
        amps_data = self.network_driver.plug_vip(db_lb,
                                                 db_lb.vip)
        return [amp.to_dict() for amp in amps_data]

    def revert(self, result, loadbalancer, *args, **kwargs):
        """Handle a failure to plumb a vip."""

        if isinstance(result, failure.Failure):
            return
        LOG.warning("Unable to plug VIP for loadbalancer id %s",
                    loadbalancer[constants.LOADBALANCER_ID])

        db_lb = self.loadbalancer_repo.get(
            db_apis.get_session(),
            id=loadbalancer[constants.LOADBALANCER_ID])
        try:
            # Make sure we have the current port IDs for cleanup
            for amp_data in result:
                for amphora in filter(
                        # pylint: disable=cell-var-from-loop
                        lambda amp: amp.id == amp_data['id'],
                        db_lb.amphorae):
                    amphora.vrrp_port_id = amp_data['vrrp_port_id']
                    amphora.ha_port_id = amp_data['ha_port_id']

            self.network_driver.unplug_vip(db_lb, db_lb.vip)
        except Exception as e:
            LOG.error("Failed to unplug VIP.  Resources may still "
                      "be in use from vip: %(vip)s due to error: %(except)s",
                      {'vip': loadbalancer['vip_address'], 'except': str(e)})


class UpdateVIPSecurityGroup(BaseNetworkTask):
    """Task to setup SG for LB."""

    def execute(self, loadbalancer_id):
        """Task to setup SG for LB."""

        LOG.debug("Setting up VIP SG for load balancer id: %s",
                  loadbalancer_id)
        db_lb = self.loadbalancer_repo.get(
            db_apis.get_session(), id=loadbalancer_id)

        sg_id = self.network_driver.update_vip_sg(db_lb, db_lb.vip)
        LOG.info("Set up VIP SG %s for load balancer %s complete",
                 sg_id if sg_id else "None", loadbalancer_id)
        return sg_id


class GetSubnetFromVIP(BaseNetworkTask):
    """Task to plumb a VIP."""

    def execute(self, loadbalancer):
        """Plumb a vip to an amphora."""

        LOG.debug("Getting subnet for LB: %s",
                  loadbalancer[constants.LOADBALANCER_ID])

        subnet = self.network_driver.get_subnet(loadbalancer['vip_subnet_id'])
        LOG.info("Got subnet %s for load balancer %s",
                 loadbalancer['vip_subnet_id'] if subnet else "None",
                 loadbalancer[constants.LOADBALANCER_ID])
        return subnet.to_dict()


class PlugVIPAmphora(BaseNetworkTask):
    """Task to plumb a VIP."""

    def execute(self, loadbalancer, amphora, subnet):
        """Plumb a vip to an amphora."""

        LOG.debug("Plumbing VIP for amphora id: %s",
                  amphora.get(constants.ID))
        db_amp = self.amphora_repo.get(db_apis.get_session(),
                                       id=amphora.get(constants.ID))
        db_subnet = self.network_driver.get_subnet(subnet[constants.ID])
        db_lb = self.loadbalancer_repo.get(
            db_apis.get_session(), id=loadbalancer[constants.LOADBALANCER_ID])
        amp_data = self.network_driver.plug_aap_port(
            db_lb, db_lb.vip, db_amp, db_subnet)
        return amp_data.to_dict()

    def revert(self, result, loadbalancer, amphora, subnet, *args, **kwargs):
        """Handle a failure to plumb a vip."""
        if isinstance(result, failure.Failure):
            return
        LOG.warning("Unable to plug VIP for amphora id %s "
                    "load balancer id %s",
                    amphora.get(constants.ID),
                    loadbalancer[constants.LOADBALANCER_ID])

        try:
            db_amp = self.amphora_repo.get(db_apis.get_session(),
                                           id=amphora.get(constants.ID))
            db_amp.vrrp_port_id = result[constants.VRRP_PORT_ID]
            db_amp.ha_port_id = result[constants.HA_PORT_ID]
            db_subnet = self.network_driver.get_subnet(subnet[constants.ID])
            db_lb = self.loadbalancer_repo.get(
                db_apis.get_session(),
                id=loadbalancer[constants.LOADBALANCER_ID])

            self.network_driver.unplug_aap_port(db_lb.vip,
                                                db_amp, db_subnet)
        except Exception as e:
            LOG.error('Failed to unplug AAP port. Resources may still be in '
                      'use for VIP: %s due to error: %s', db_lb.vip, str(e))


class UnplugVIP(BaseNetworkTask):
    """Task to unplug the vip."""

    def execute(self, loadbalancer):
        """Unplug the vip."""

        LOG.debug("Unplug vip on amphora")
        try:
            db_lb = self.loadbalancer_repo.get(
                db_apis.get_session(),
                id=loadbalancer[constants.LOADBALANCER_ID])
            self.network_driver.unplug_vip(db_lb, db_lb.vip)
        except Exception:
            LOG.exception("Unable to unplug vip from load balancer %s",
                          loadbalancer[constants.LOADBALANCER_ID])


class AllocateVIP(BaseNetworkTask):
    """Task to allocate a VIP."""

    def execute(self, loadbalancer):
        """Allocate a vip to the loadbalancer."""

        LOG.debug("Allocating vip with port id %s, subnet id %s, "
                  "ip address %s for load balancer %s",
                  loadbalancer[constants.VIP_PORT_ID],
                  loadbalancer[constants.VIP_SUBNET_ID],
                  loadbalancer[constants.VIP_ADDRESS],
                  loadbalancer[constants.LOADBALANCER_ID])
        db_lb = self.loadbalancer_repo.get(
            db_apis.get_session(), id=loadbalancer[constants.LOADBALANCER_ID])
        vip = self.network_driver.allocate_vip(db_lb)
        LOG.info("Allocated vip with port id %s, subnet id %s, ip address %s "
                 "for load balancer %s",
                 loadbalancer[constants.VIP_PORT_ID],
                 loadbalancer[constants.VIP_SUBNET_ID],
                 loadbalancer[constants.VIP_ADDRESS],
                 loadbalancer[constants.LOADBALANCER_ID])
        return vip.to_dict()

    def revert(self, result, loadbalancer, *args, **kwargs):
        """Handle a failure to allocate vip."""

        if isinstance(result, failure.Failure):
            LOG.exception("Unable to allocate VIP")
            return
        vip = data_models.Vip(**result)
        LOG.warning("Deallocating vip %s", vip.ip_address)
        try:
            self.network_driver.deallocate_vip(vip)
        except Exception as e:
            LOG.error("Failed to deallocate VIP.  Resources may still "
                      "be in use from vip: %(vip)s due to error: %(except)s",
                      {'vip': vip.ip_address, 'except': str(e)})


class AllocateVIPforFailover(AllocateVIP):
    """Task to allocate/validate the VIP for a failover flow."""

    def revert(self, result, loadbalancer, *args, **kwargs):
        """Handle a failure to allocate vip."""

        if isinstance(result, failure.Failure):
            LOG.exception("Unable to allocate VIP")
            return
        vip = data_models.Vip(**result)
        LOG.info("Failover revert is not deallocating vip %s because this is "
                 "a failover.", vip.ip_address)


class DeallocateVIP(BaseNetworkTask):
    """Task to deallocate a VIP."""

    def execute(self, loadbalancer):
        """Deallocate a VIP."""

        LOG.debug("Deallocating a VIP %s", loadbalancer[constants.VIP_ADDRESS])

        # NOTE(blogan): this is kind of ugly but sufficient for now.  Drivers
        # will need access to the load balancer that the vip is/was attached
        # to.  However the data model serialization for the vip does not give a
        # backref to the loadbalancer if accessed through the loadbalancer.
        db_lb = self.loadbalancer_repo.get(
            db_apis.get_session(), id=loadbalancer[constants.LOADBALANCER_ID])
        vip = db_lb.vip
        vip.load_balancer = db_lb
        self.network_driver.deallocate_vip(vip)


class UpdateVIP(BaseNetworkTask):
    """Task to update a VIP."""

    def execute(self, listeners):
        loadbalancer = self.loadbalancer_repo.get(
            db_apis.get_session(), id=listeners[0][constants.LOADBALANCER_ID])

        LOG.debug("Updating VIP of load_balancer %s.", loadbalancer.id)

        self.network_driver.update_vip(loadbalancer)


class UpdateVIPForDelete(BaseNetworkTask):
    """Task to update a VIP for listener delete flows."""

    def execute(self, loadbalancer_id):
        loadbalancer = self.loadbalancer_repo.get(
            db_apis.get_session(), id=loadbalancer_id)
        LOG.debug("Updating VIP for listener delete on load_balancer %s.",
                  loadbalancer.id)
        self.network_driver.update_vip(loadbalancer, for_delete=True)


class GetAmphoraNetworkConfigs(BaseNetworkTask):
    """Task to retrieve amphora network details."""

    def execute(self, loadbalancer, amphora=None):
        LOG.debug("Retrieving vip network details.")
        db_amp = self.amphora_repo.get(db_apis.get_session(),
                                       id=amphora.get(constants.ID))
        db_lb = self.loadbalancer_repo.get(
            db_apis.get_session(), id=loadbalancer[constants.LOADBALANCER_ID])
        db_configs = self.network_driver.get_network_configs(
            db_lb, amphora=db_amp)
        provider_dict = {}
        for amp_id, amp_conf in db_configs.items():
            provider_dict[amp_id] = amp_conf.to_dict(recurse=True)
        return provider_dict


class GetAmphoraNetworkConfigsByID(BaseNetworkTask):
    """Task to retrieve amphora network details."""

    def execute(self, loadbalancer_id, amphora_id=None):
        LOG.debug("Retrieving vip network details.")
        loadbalancer = self.loadbalancer_repo.get(db_apis.get_session(),
                                                  id=loadbalancer_id)
        amphora = self.amphora_repo.get(db_apis.get_session(), id=amphora_id)
        db_configs = self.network_driver.get_network_configs(loadbalancer,
                                                             amphora=amphora)
        provider_dict = {}
        for amp_id, amp_conf in db_configs.items():
            provider_dict[amp_id] = amp_conf.to_dict(recurse=True)
        return provider_dict


class GetAmphoraeNetworkConfigs(BaseNetworkTask):
    """Task to retrieve amphorae network details."""

    def execute(self, loadbalancer_id):
        LOG.debug("Retrieving vip network details.")
        db_lb = self.loadbalancer_repo.get(
            db_apis.get_session(), id=loadbalancer_id)
        db_configs = self.network_driver.get_network_configs(db_lb)
        provider_dict = {}
        for amp_id, amp_conf in db_configs.items():
            provider_dict[amp_id] = amp_conf.to_dict(recurse=True)
        return provider_dict


class FailoverPreparationForAmphora(BaseNetworkTask):
    """Task to prepare an amphora for failover."""

    def execute(self, amphora):
        db_amp = self.amphora_repo.get(db_apis.get_session(),
                                       id=amphora[constants.ID])
        LOG.debug("Prepare amphora %s for failover.", amphora[constants.ID])

        self.network_driver.failover_preparation(db_amp)


class RetrievePortIDsOnAmphoraExceptLBNetwork(BaseNetworkTask):
    """Task retrieving all the port ids on an amphora, except lb network."""

    def execute(self, amphora):
        LOG.debug("Retrieve all but the lb network port id on amphora %s.",
                  amphora[constants.ID])

        interfaces = self.network_driver.get_plugged_networks(
            compute_id=amphora[constants.COMPUTE_ID])

        ports = []
        for interface_ in interfaces:
            if interface_.port_id not in ports:
                port = self.network_driver.get_port(port_id=interface_.port_id)
                ips = port.fixed_ips
                lb_network = False
                for ip in ips:
                    if ip.ip_address == amphora[constants.LB_NETWORK_IP]:
                        lb_network = True
                if not lb_network:
                    ports.append(port)

        return ports


class PlugPorts(BaseNetworkTask):
    """Task to plug neutron ports into a compute instance."""

    def execute(self, amphora, ports):
        db_amp = self.amphora_repo.get(db_apis.get_session(),
                                       id=amphora[constants.ID])
        for port in ports:
            LOG.debug('Plugging port ID: %(port_id)s into compute instance: '
                      '%(compute_id)s.',
                      {constants.PORT_ID: port.id,
                       constants.COMPUTE_ID: amphora[constants.COMPUTE_ID]})
            self.network_driver.plug_port(db_amp, port)


class ApplyQos(BaseNetworkTask):
    """Apply Quality of Services to the VIP"""

    def _apply_qos_on_vrrp_ports(self, loadbalancer, amps_data, qos_policy_id,
                                 is_revert=False, request_qos_id=None):
        """Call network driver to apply QoS Policy on the vrrp ports."""

        if not amps_data:
            db_lb = self.loadbalancer_repo.get(
                db_apis.get_session(),
                id=loadbalancer[constants.LOADBALANCER_ID])
            amps_data = db_lb.amphorae

        apply_qos = ApplyQosAmphora()
        for amp_data in amps_data:
            apply_qos._apply_qos_on_vrrp_port(loadbalancer, amp_data.to_dict(),
                                              qos_policy_id)

    def execute(self, loadbalancer, amps_data=None, update_dict=None):
        """Apply qos policy on the vrrp ports which are related with vip."""
        db_lb = self.loadbalancer_repo.get(
            db_apis.get_session(),
            id=loadbalancer[constants.LOADBALANCER_ID])

        qos_policy_id = db_lb.vip.qos_policy_id
        if not qos_policy_id and (
            not update_dict or (
                'vip' not in update_dict or
                'qos_policy_id' not in update_dict[constants.VIP])):
            return
        if update_dict and update_dict.get(constants.VIP):
            vip_dict = update_dict[constants.VIP]
            if vip_dict.get(constants.QOS_POLICY_ID):
                qos_policy_id = vip_dict[constants.QOS_POLICY_ID]

        self._apply_qos_on_vrrp_ports(loadbalancer, amps_data, qos_policy_id)

    def revert(self, result, loadbalancer, amps_data=None, update_dict=None,
               *args, **kwargs):
        """Handle a failure to apply QoS to VIP"""

        request_qos_id = loadbalancer['vip_qos_policy_id']
        orig_lb = self.task_utils.get_current_loadbalancer_from_db(
            loadbalancer[constants.LOADBALANCER_ID])
        orig_qos_id = orig_lb.vip.qos_policy_id
        if request_qos_id != orig_qos_id:
            self._apply_qos_on_vrrp_ports(loadbalancer, amps_data, orig_qos_id,
                                          is_revert=True,
                                          request_qos_id=request_qos_id)


class ApplyQosAmphora(BaseNetworkTask):
    """Apply Quality of Services to the VIP"""

    def _apply_qos_on_vrrp_port(self, loadbalancer, amp_data, qos_policy_id,
                                is_revert=False, request_qos_id=None):
        """Call network driver to apply QoS Policy on the vrrp ports."""
        try:
            self.network_driver.apply_qos_on_port(
                qos_policy_id,
                amp_data[constants.VRRP_PORT_ID])
        except Exception:
            if not is_revert:
                raise
            LOG.warning('Failed to undo qos policy %(qos_id)s '
                        'on vrrp port: %(port)s from '
                        'amphorae: %(amp)s',
                        {'qos_id': request_qos_id,
                         'port': amp_data[constants.VRRP_PORT_ID],
                         'amp': [amp.get(constants.ID) for amp in amp_data]})

    def execute(self, loadbalancer, amp_data=None, update_dict=None):
        """Apply qos policy on the vrrp ports which are related with vip."""
        qos_policy_id = loadbalancer['vip_qos_policy_id']
        if not qos_policy_id and (
            update_dict and (
                'vip' not in update_dict or
                'qos_policy_id' not in update_dict[constants.VIP])):
            return
        self._apply_qos_on_vrrp_port(loadbalancer, amp_data, qos_policy_id)

    def revert(self, result, loadbalancer, amp_data=None, update_dict=None,
               *args, **kwargs):
        """Handle a failure to apply QoS to VIP"""
        try:
            request_qos_id = loadbalancer['vip_qos_policy_id']
            orig_lb = self.task_utils.get_current_loadbalancer_from_db(
                loadbalancer[constants.LOADBALANCER_ID])
            orig_qos_id = orig_lb.vip.qos_policy_id
            if request_qos_id != orig_qos_id:
                self._apply_qos_on_vrrp_port(loadbalancer, amp_data,
                                             orig_qos_id, is_revert=True,
                                             request_qos_id=request_qos_id)
        except Exception as e:
            LOG.error('Failed to remove QoS policy: %s from port: %s due '
                      'to error: %s', orig_qos_id,
                      amp_data[constants.VRRP_PORT_ID], str(e))


class DeletePort(BaseNetworkTask):
    """Task to delete a network port."""

    @tenacity.retry(retry=tenacity.retry_if_exception_type(),
                    stop=tenacity.stop_after_attempt(
                        CONF.networking.max_retries),
                    wait=tenacity.wait_exponential(
                        multiplier=CONF.networking.retry_backoff,
                        min=CONF.networking.retry_interval,
                        max=CONF.networking.retry_max), reraise=True)
    def execute(self, port_id, passive_failure=False):
        """Delete the network port."""
        if port_id is None:
            return
        if self.execute.retry.statistics.get(constants.ATTEMPT_NUMBER, 1) == 1:
            LOG.debug("Deleting network port %s", port_id)
        else:
            LOG.warning('Retrying network port %s delete attempt %s of %s.',
                        port_id,
                        self.execute.retry.statistics[
                            constants.ATTEMPT_NUMBER],
                        self.execute.retry.stop.max_attempt_number)
        # Let the Taskflow engine know we are working and alive
        # Don't use get with a default for 'attempt_number', we need to fail
        # if that number is missing.
        self.update_progress(
            self.execute.retry.statistics[constants.ATTEMPT_NUMBER] /
            self.execute.retry.stop.max_attempt_number)
        try:
            self.network_driver.delete_port(port_id)
        except Exception:
            if (self.execute.retry.statistics[constants.ATTEMPT_NUMBER] !=
                    self.execute.retry.stop.max_attempt_number):
                LOG.warning('Network port delete for port id: %s failed. '
                            'Retrying.', port_id)
                raise
            if passive_failure:
                LOG.exception('Network port delete for port ID: %s failed. '
                              'This resource will be abandoned and should '
                              'manually be cleaned up once the '
                              'network service is functional.', port_id)
                # Let's at least attempt to disable it so if the instance
                # comes back from the dead it doesn't conflict with anything.
                try:
                    self.network_driver.admin_down_port(port_id)
                    LOG.info('Successfully disabled (admin down) network port '
                             '%s that failed to delete.', port_id)
                except Exception:
                    LOG.warning('Attempt to disable (admin down) network port '
                                '%s failed. The network service has failed. '
                                'Continuing.', port_id)
            else:
                LOG.exception('Network port delete for port ID: %s failed. '
                              'The network service has failed. '
                              'Aborting and reverting.', port_id)
                raise


class CreateVIPBasePort(BaseNetworkTask):
    """Task to create the VIP base port for an amphora."""

    @tenacity.retry(retry=tenacity.retry_if_exception_type(),
                    stop=tenacity.stop_after_attempt(
                        CONF.networking.max_retries),
                    wait=tenacity.wait_exponential(
                        multiplier=CONF.networking.retry_backoff,
                        min=CONF.networking.retry_interval,
                        max=CONF.networking.retry_max), reraise=True)
    def execute(self, vip, vip_sg_id, amphora_id):
        port_name = constants.AMP_BASE_PORT_PREFIX + amphora_id
        fixed_ips = [{constants.SUBNET_ID: vip[constants.SUBNET_ID]}]
        sg_id = []
        if vip_sg_id:
            sg_id = [vip_sg_id]
        port = self.network_driver.create_port(
            vip[constants.NETWORK_ID], name=port_name, fixed_ips=fixed_ips,
            secondary_ips=[vip[constants.IP_ADDRESS]],
            security_group_ids=sg_id,
            qos_policy_id=vip[constants.QOS_POLICY_ID])
        LOG.info('Created port %s with ID %s for amphora %s',
                 port_name, port.id, amphora_id)
        return port.to_dict(recurse=True)

    def revert(self, result, vip, vip_sg_id, amphora_id, *args, **kwargs):
        if isinstance(result, failure.Failure):
            return
        try:
            port_name = constants.AMP_BASE_PORT_PREFIX + amphora_id
            for port in result:
                self.network_driver.delete_port(port.id)
                LOG.info('Deleted port %s with ID %s for amphora %s due to a '
                         'revert.', port_name, port.id, amphora_id)
        except Exception as e:
            LOG.error('Failed to delete port %s. Resources may still be in '
                      'use for a port intended for amphora %s due to error '
                      '%s. Search for a port named %s',
                      result, amphora_id, str(e), port_name)


class AdminDownPort(BaseNetworkTask):

    def execute(self, port_id):
        try:
            self.network_driver.set_port_admin_state_up(port_id, False)
        except base.PortNotFound:
            return
        for i in range(CONF.networking.max_retries):
            port = self.network_driver.get_port(port_id)
            if port.status == constants.DOWN:
                LOG.debug('Disabled port: %s', port_id)
                return
            LOG.debug('Port %s is %s instead of DOWN, waiting.',
                      port_id, port.status)
            time.sleep(CONF.networking.retry_interval)
        LOG.error('Port %s failed to go DOWN. Port status is still %s. '
                  'Ignoring and continuing.', port_id, port.status)

    def revert(self, result, port_id, *args, **kwargs):
        if isinstance(result, failure.Failure):
            return
        try:
            self.network_driver.set_port_admin_state_up(port_id, True)
        except Exception as e:
            LOG.error('Failed to bring port %s admin up on revert due to: %s.',
                      port_id, str(e))


class GetVIPSecurityGroupID(BaseNetworkTask):

    def execute(self, loadbalancer_id):
        sg_name = utils.get_vip_security_group_name(loadbalancer_id)
        try:
            security_group = self.network_driver.get_security_group(sg_name)
            if security_group:
                return security_group.id
        except base.SecurityGroupNotFound:
            with excutils.save_and_reraise_exception() as ctxt:
                if self.network_driver.sec_grp_enabled:
                    LOG.error('VIP security group %s was not found.', sg_name)
                else:
                    ctxt.reraise = False
        return None



#    Copyright 2014 Rackspace
#    Copyright 2016 Blue Box, an IBM Company
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_utils import excutils
from pecan import expose as pecan_expose
from pecan import request as pecan_request
from sqlalchemy.orm import exc as sa_exception
from wsme import types as wtypes
from wsmeext import pecan as wsme_pecan

from octavia.api.v2.controllers import base
from octavia.api.v2.types import amphora as amp_types
from octavia.common import constants
from octavia.common import exceptions
from octavia.common import rpc
from octavia.db import api as db_api

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class AmphoraController(base.BaseController):
    RBAC_TYPE = constants.RBAC_AMPHORA

    def __init__(self):
        super().__init__()
        topic = cfg.CONF.oslo_messaging.topic
        self.target = messaging.Target(
            namespace=constants.RPC_NAMESPACE_CONTROLLER_AGENT,
            topic=topic, version="1.0", fanout=False)
        self.client = rpc.get_client(self.target)

    @wsme_pecan.wsexpose(amp_types.AmphoraRootResponse, wtypes.text,
                         [wtypes.text], ignore_extra_args=True)
    def get_one(self, id, fields=None):
        """Gets a single amphora's details."""
        context = pecan_request.context.get('octavia_context')
        db_amp = self._get_db_amp(context.session, id, show_deleted=False)

        self._auth_validate_action(context, context.project_id,
                                   constants.RBAC_GET_ONE)

        result = self._convert_db_to_type(
            db_amp, amp_types.AmphoraResponse)
        if fields is not None:
            result = self._filter_fields([result], fields)[0]
        return amp_types.AmphoraRootResponse(amphora=result)

    @wsme_pecan.wsexpose(amp_types.AmphoraeRootResponse, [wtypes.text],
                         ignore_extra_args=True)
    def get_all(self, fields=None):
        """Gets all amphorae."""
        pcontext = pecan_request.context
        context = pcontext.get('octavia_context')

        self._auth_validate_action(context, context.project_id,
                                   constants.RBAC_GET_ALL)

        db_amp, links = self.repositories.amphora.get_all_API_list(
            context.session, show_deleted=False,
            pagination_helper=pcontext.get(constants.PAGINATION_HELPER))
        result = self._convert_db_to_type(
            db_amp, [amp_types.AmphoraResponse])
        if fields is not None:
            result = self._filter_fields(result, fields)
        return amp_types.AmphoraeRootResponse(
            amphorae=result, amphorae_links=links)

    @wsme_pecan.wsexpose(None, wtypes.text, status_code=204)
    def delete(self, id):
        """Deletes an amphora."""
        context = pecan_request.context.get('octavia_context')

        self._auth_validate_action(context, context.project_id,
                                   constants.RBAC_DELETE)

        with db_api.get_lock_session() as lock_session:
            try:
                self.repositories.amphora.test_and_set_status_for_delete(
                    lock_session, id)
            except sa_exception.NoResultFound as e:
                raise exceptions.NotFound(resource='Amphora', id=id) from e

            LOG.info("Sending delete amphora %s to the queue.", id)
            payload = {constants.AMPHORA_ID: id}
            self.client.cast({}, 'delete_amphora', **payload)

    @pecan_expose()
    def _lookup(self, amphora_id, *remainder):
        """Overridden pecan _lookup method for custom routing.

        Currently it checks if this was a failover request and routes
        the request to the FailoverController.
        """
        if amphora_id and remainder:
            controller = remainder[0]
            remainder = remainder[1:]
            if controller == 'config':
                return AmphoraUpdateController(amp_id=amphora_id), remainder
            if controller == 'failover':
                return FailoverController(amp_id=amphora_id), remainder
            if controller == 'stats':
                return AmphoraStatsController(amp_id=amphora_id), remainder
        return None


class FailoverController(base.BaseController):
    RBAC_TYPE = constants.RBAC_AMPHORA

    def __init__(self, amp_id):
        super().__init__()
        if CONF.api_settings.default_provider_driver == constants.AMPHORAV1:
            topic = cfg.CONF.oslo_messaging.topic
            version = "1.0"
        else:
            topic = constants.TOPIC_AMPHORA_V2
            version = "2.0"
        self.target = messaging.Target(
            namespace=constants.RPC_NAMESPACE_CONTROLLER_AGENT,
            topic=topic, version=version, fanout=False)
        self.client = rpc.get_client(self.target)
        self.amp_id = amp_id

    @wsme_pecan.wsexpose(None, wtypes.text, status_code=202)
    def put(self):
        """Fails over an amphora"""
        pcontext = pecan_request.context
        context = pcontext.get('octavia_context')
        db_amp = self._get_db_amp(context.session, self.amp_id,
                                  show_deleted=False)

        self._auth_validate_action(
            context, db_amp.load_balancer.project_id,
            constants.RBAC_PUT_FAILOVER)

        self.repositories.load_balancer.test_and_set_provisioning_status(
            context.session, db_amp.load_balancer_id,
            status=constants.PENDING_UPDATE, raise_exception=True)

        try:
            LOG.info("Sending failover request for amphora %s to the queue",
                     self.amp_id)
            payload = {constants.AMPHORA_ID: db_amp.id}
            self.client.cast({}, 'failover_amphora', **payload)
        except Exception:
            with excutils.save_and_reraise_exception(reraise=False):
                self.repositories.load_balancer.update(
                    context.session, db_amp.load_balancer.id,
                    provisioning_status=constants.ERROR)


class AmphoraUpdateController(base.BaseController):
    RBAC_TYPE = constants.RBAC_AMPHORA

    def __init__(self, amp_id):
        super().__init__()

        if CONF.api_settings.default_provider_driver == constants.AMPHORAV1:
            topic = cfg.CONF.oslo_messaging.topic
            version = "1.0"
        else:
            topic = constants.TOPIC_AMPHORA_V2
            version = "2.0"
        self.transport = messaging.get_rpc_transport(cfg.CONF)
        self.target = messaging.Target(
            namespace=constants.RPC_NAMESPACE_CONTROLLER_AGENT,
            topic=topic, version=version, fanout=False)
        self.client = messaging.RPCClient(self.transport, target=self.target)
        self.amp_id = amp_id

    @wsme_pecan.wsexpose(None, wtypes.text, status_code=202)
    def put(self):
        """Update amphora agent configuration"""
        pcontext = pecan_request.context
        context = pcontext.get('octavia_context')
        db_amp = self._get_db_amp(context.session, self.amp_id,
                                  show_deleted=False)

        self._auth_validate_action(
            context, db_amp.load_balancer.project_id,
            constants.RBAC_PUT_CONFIG)

        try:
            LOG.info("Sending amphora agent update request for amphora %s to "
                     "the queue.", self.amp_id)
            payload = {constants.AMPHORA_ID: db_amp.id}
            self.client.cast({}, 'update_amphora_agent_config', **payload)
        except Exception:
            with excutils.save_and_reraise_exception(reraise=True):
                LOG.error("Unable to send amphora agent update request for "
                          "amphora %s to the queue.", self.amp_id)


class AmphoraStatsController(base.BaseController):
    RBAC_TYPE = constants.RBAC_AMPHORA

    def __init__(self, amp_id):
        super().__init__()
        self.amp_id = amp_id

    @wsme_pecan.wsexpose(amp_types.StatisticsRootResponse, wtypes.text,
                         status_code=200)
    def get(self):
        context = pecan_request.context.get('octavia_context')

        self._auth_validate_action(context, context.project_id,
                                   constants.RBAC_GET_STATS)

        stats = self.repositories.get_amphora_stats(context.session,
                                                    self.amp_id)
        if not stats:
            raise exceptions.NotFound(resource='Amphora stats for',
                                      id=self.amp_id)

        wsme_stats = []
        for stat in stats:
            wsme_stats.append(amp_types.AmphoraStatisticsResponse(**stat))
        return amp_types.StatisticsRootResponse(amphora_stats=wsme_stats)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
from unittest import mock

from taskflow.patterns import linear_flow as flow

from octavia.common import constants
from octavia.controller.worker.v2.flows import member_flows
import octavia.tests.unit.base as base


# NOTE: We patch the get_network_driver for all the calls so we don't
# inadvertently make real calls.
@mock.patch('octavia.common.utils.get_network_driver')
class TestMemberFlows(base.TestCase):

    def setUp(self):
        self.MemberFlow = member_flows.MemberFlows()

        super().setUp()

    def test_get_create_member_flow(self, mock_get_net_driver):

        member_flow = self.MemberFlow.get_create_member_flow()

        self.assertIsInstance(member_flow, flow.Flow)

        self.assertIn(constants.LISTENERS, member_flow.requires)
        self.assertIn(constants.LOADBALANCER, member_flow.requires)
        self.assertIn(constants.LOADBALANCER_ID, member_flow.requires)
        self.assertIn(constants.POOL_ID, member_flow.requires)
        self.assertIn(constants.MEMBER, member_flow.requires)
        self.assertIn(constants.AVAILABILITY_ZONE, member_flow.requires)

        self.assertIn(constants.DELTAS, member_flow.provides)
        self.assertIn(constants.ADDED_PORTS, member_flow.provides)

        self.assertEqual(6, len(member_flow.requires))
        self.assertEqual(2, len(member_flow.provides))

    def test_get_delete_member_flow(self, mock_get_net_driver):

        member_flow = self.MemberFlow.get_delete_member_flow()

        self.assertIsInstance(member_flow, flow.Flow)

        self.assertIn(constants.MEMBER, member_flow.requires)
        self.assertIn(constants.LOADBALANCER, member_flow.requires)
        self.assertIn(constants.LISTENERS, member_flow.requires)
        self.assertIn(constants.LOADBALANCER_ID, member_flow.requires)
        self.assertIn(constants.POOL_ID, member_flow.requires)
        self.assertIn(constants.PROJECT_ID, member_flow.requires)

        self.assertEqual(6, len(member_flow.requires))
        self.assertEqual(0, len(member_flow.provides))

    def test_get_update_member_flow(self, mock_get_net_driver):

        member_flow = self.MemberFlow.get_update_member_flow()

        self.assertIsInstance(member_flow, flow.Flow)

        self.assertIn(constants.MEMBER, member_flow.requires)
        self.assertIn(constants.LISTENERS, member_flow.requires)
        self.assertIn(constants.LOADBALANCER, member_flow.requires)
        self.assertIn(constants.LOADBALANCER_ID, member_flow.requires)
        self.assertIn(constants.POOL_ID, member_flow.requires)
        self.assertIn(constants.UPDATE_DICT, member_flow.requires)

        self.assertEqual(6, len(member_flow.requires))
        self.assertEqual(0, len(member_flow.provides))

    def test_get_batch_update_members_flow(self, mock_get_net_driver):

        member_flow = self.MemberFlow.get_batch_update_members_flow(
            [], [], [])

        self.assertIsInstance(member_flow, flow.Flow)

        self.assertIn(constants.LISTENERS, member_flow.requires)
        self.assertIn(constants.LOADBALANCER, member_flow.requires)
        self.assertIn(constants.LOADBALANCER_ID, member_flow.requires)
        self.assertIn(constants.POOL_ID, member_flow.requires)
        self.assertIn(constants.AVAILABILITY_ZONE, member_flow.requires)

        self.assertIn(constants.DELTAS, member_flow.provides)
        self.assertIn(constants.ADDED_PORTS, member_flow.provides)

        self.assertEqual(5, len(member_flow.requires))
        self.assertEqual(2, len(member_flow.provides))



# Copyright 2021 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

from octavia.amphorae.backends.utils import interface_file
from octavia.cmd import interface
from octavia.tests.unit import base


class TestInterfaceCMD(base.TestCase):

    def setUp(self):
        super().setUp()

        self.interface1 = interface_file.InterfaceFile("eth1")
        self.interface2 = interface_file.InterfaceFile("eth2")

    def test_interfaces_find(self):
        controller = mock.Mock()
        controller.list = mock.Mock()
        controller.list.return_value = {
            "eth1": self.interface1,
            "eth2": self.interface2
        }

        ret = interface.interfaces_find(controller, "eth2")
        self.assertCountEqual([self.interface2], ret)
        controller.list.assert_called_once()

    def test_interfaces_find_all(self):
        controller = mock.Mock()
        controller.list = mock.Mock()
        controller.list.return_value = {
            "eth1": self.interface1,
            "eth2": self.interface2
        }

        ret = interface.interfaces_find(controller, "all")
        self.assertCountEqual([self.interface1, self.interface2], ret)
        controller.list.assert_called_once()

    def test_interfaces_find_all_empty(self):
        controller = mock.Mock()
        controller.list = mock.Mock()
        controller.list.return_value = {}

        ret = interface.interfaces_find(controller, "all")
        self.assertEqual(0, len(ret))
        controller.list.assert_called_once()

    def test_interfaces_find_not_found(self):
        controller = mock.Mock()
        controller.list = mock.Mock()
        controller.list.return_value = {
            "eth1": self.interface1,
            "eth2": self.interface2
        }

        self.assertRaisesRegex(
            interface.InterfaceException,
            "Could not find interface 'eth3'.",
            interface.interfaces_find,
            controller, "eth3")
        controller.list.assert_called_once()

    def test_interfaces_update(self):
        action_fn = mock.Mock()
        action_str = mock.Mock()
        interfaces = [self.interface1, self.interface2]

        interface.interfaces_update(interfaces, action_fn, action_str)
        self.assertEqual(2, len(action_fn.mock_calls))
        action_fn.assert_called_with(self.interface2)

    def test_interfaces_update_with_errors(self):
        action_fn = mock.Mock()
        action_str = mock.Mock()
        interfaces = [self.interface1, self.interface2]
        action_fn.side_effect = [None, Exception("error msg")]

        self.assertRaisesRegex(
            interface.InterfaceException,
            "Could not configure interface:.*eth2.*error msg",
            interface.interfaces_update,
            interfaces, action_fn, action_str)
        self.assertEqual(2, len(action_fn.mock_calls))

    @mock.patch("octavia.amphorae.backends.utils.interface."
                "InterfaceController")
    @mock.patch("octavia.cmd.interface.interfaces_find")
    @mock.patch("octavia.cmd.interface.interfaces_update")
    def test_interface_cmd(self, mock_interfaces_update,
                           mock_interfaces_find, mock_controller):
        controller = mock.Mock()
        controller.up = mock.Mock()
        controller.down = mock.Mock()
        mock_controller.return_value = controller
        mock_interfaces_find.return_value = [self.interface1]

        interface.interface_cmd("eth1", "up")

        mock_interfaces_find.assert_called_once_with(
            controller, "eth1")
        mock_interfaces_update.assert_called_once_with(
            [self.interface1], mock_controller.return_value.up, "up")

        mock_interfaces_find.reset_mock()
        mock_interfaces_update.reset_mock()

        mock_interfaces_find.return_value = [self.interface2]

        interface.interface_cmd("eth2", "down")

        mock_interfaces_find.assert_called_once_with(
            controller, "eth2")
        mock_interfaces_update.assert_called_once_with(
            [self.interface2], mock_controller.return_value.down, "down")

    @mock.patch("octavia.amphorae.backends.utils.interface."
                "InterfaceController")
    def test_interface_cmd_invalid_action(self, mock_controller):
        self.assertRaisesRegex(
            interface.InterfaceException,
            "Unknown action.*invalidaction",
            interface.interface_cmd,
            "eth1", "invalidaction")



# Copyright 2016 Blue Box, an IBM Company
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Several handy validation functions that go beyond simple type checking.
Defined here so these can also be used at deeper levels than the API.
"""


import ipaddress
import re

import netaddr
from oslo_config import cfg
import rfc3986
from wsme import types as wtypes

from octavia.common import constants
from octavia.common import exceptions
from octavia.common import utils
from octavia.i18n import _

CONF = cfg.CONF


def url(url, require_scheme=True):
    """Raises an error if the url doesn't look like a URL."""
    try:
        if not rfc3986.is_valid_uri(url, require_scheme=require_scheme):
            raise exceptions.InvalidURL(url=url)
        p_url = rfc3986.urlparse(rfc3986.normalize_uri(url))
        if require_scheme:
            if p_url.scheme not in ('http', 'https'):
                raise exceptions.InvalidURL(url=url)
    except Exception as e:
        raise exceptions.InvalidURL(url=url) from e
    return True


def url_path(url_path):
    """Raises an error if the url_path doesn't look like a URL Path."""
    try:
        p_url = rfc3986.urlparse(rfc3986.normalize_uri(url_path))

        invalid_path = (
            p_url.scheme or p_url.userinfo or p_url.host or
            p_url.port or
            p_url.path is None or
            not p_url.path.startswith('/')
        )

        if invalid_path:
            raise exceptions.InvalidURLPath(url_path=url_path)
    except Exception as e:
        raise exceptions.InvalidURLPath(url_path=url_path) from e
    return True


def header_name(header, what=None):
    """Raises an error if header does not look like an HTML header name."""
    p = re.compile(constants.HTTP_HEADER_NAME_REGEX)
    if not p.match(header):
        raise exceptions.InvalidString(what=what)
    return True


def cookie_value_string(value, what=None):
    """Raises an error if the value string contains invalid characters."""
    p = re.compile(constants.HTTP_COOKIE_VALUE_REGEX)
    if not p.match(value):
        raise exceptions.InvalidString(what=what)
    return True


def header_value_string(value, what=None):
    """Raises an error if the value string contains invalid characters."""
    p = re.compile(constants.HTTP_HEADER_VALUE_REGEX)
    q = re.compile(constants.HTTP_QUOTED_HEADER_VALUE_REGEX)
    if not p.match(value) and not q.match(value):
        raise exceptions.InvalidString(what=what)
    return True


def regex(regex):
    """Raises an error if the string given is not a valid regex."""
    try:
        re.compile(regex)
    except Exception as e:
        raise exceptions.InvalidRegex(e=str(e))
    return True


# Note that we can evaluate this outside the context of any L7 Policy because
# L7 rules must be internally consistent.
def l7rule_data(l7rule):
    """Raises an error if the l7rule given is invalid in some way."""
    if not l7rule.value:
        raise exceptions.InvalidL7Rule(msg=_('L7 rule type requires a value'))
    if l7rule.type == constants.L7RULE_TYPE_HEADER:
        if not l7rule.key:
            raise exceptions.InvalidL7Rule(msg='L7 rule type requires a key')
        header_name(l7rule.key, what='key')
        if l7rule.compare_type == constants.L7RULE_COMPARE_TYPE_REGEX:
            regex(l7rule.value)
        elif l7rule.compare_type in (
                constants.L7RULE_COMPARE_TYPE_STARTS_WITH,
                constants.L7RULE_COMPARE_TYPE_ENDS_WITH,
                constants.L7RULE_COMPARE_TYPE_CONTAINS,
                constants.L7RULE_COMPARE_TYPE_EQUAL_TO):
            header_value_string(l7rule.value, what='header value')
        else:
            raise exceptions.InvalidL7Rule(msg='invalid comparison type '
                                           'for rule type')

    elif l7rule.type == constants.L7RULE_TYPE_COOKIE:
        if not l7rule.key:
            raise exceptions.InvalidL7Rule(msg='L7 rule type requires a key')
        header_name(l7rule.key, what='key')
        if l7rule.compare_type == constants.L7RULE_COMPARE_TYPE_REGEX:
            regex(l7rule.value)
        elif l7rule.compare_type in (
                constants.L7RULE_COMPARE_TYPE_STARTS_WITH,
                constants.L7RULE_COMPARE_TYPE_ENDS_WITH,
                constants.L7RULE_COMPARE_TYPE_CONTAINS,
                constants.L7RULE_COMPARE_TYPE_EQUAL_TO):
            cookie_value_string(l7rule.value, what='cookie value')
        else:
            raise exceptions.InvalidL7Rule(msg='invalid comparison type '
                                           'for rule type')

    elif l7rule.type in (constants.L7RULE_TYPE_HOST_NAME,
                         constants.L7RULE_TYPE_PATH):
        if l7rule.compare_type in (
                constants.L7RULE_COMPARE_TYPE_STARTS_WITH,
                constants.L7RULE_COMPARE_TYPE_ENDS_WITH,
                constants.L7RULE_COMPARE_TYPE_CONTAINS,
                constants.L7RULE_COMPARE_TYPE_EQUAL_TO):
            header_value_string(l7rule.value, what='comparison value')
        elif l7rule.compare_type == constants.L7RULE_COMPARE_TYPE_REGEX:
            regex(l7rule.value)
        else:
            raise exceptions.InvalidL7Rule(msg='invalid comparison type '
                                           'for rule type')

    elif l7rule.type == constants.L7RULE_TYPE_FILE_TYPE:
        if l7rule.compare_type == constants.L7RULE_COMPARE_TYPE_REGEX:
            regex(l7rule.value)
        elif l7rule.compare_type == constants.L7RULE_COMPARE_TYPE_EQUAL_TO:
            header_value_string(l7rule.value, what='comparison value')
        else:
            raise exceptions.InvalidL7Rule(msg='invalid comparison type '
                                           'for rule type')

    elif l7rule.type in [constants.L7RULE_TYPE_SSL_CONN_HAS_CERT,
                         constants.L7RULE_TYPE_SSL_VERIFY_RESULT,
                         constants.L7RULE_TYPE_SSL_DN_FIELD]:
        validate_l7rule_ssl_types(l7rule)

    else:
        raise exceptions.InvalidL7Rule(msg='invalid rule type')
    return True


def validate_l7rule_ssl_types(l7rule):
    if not l7rule.type or l7rule.type not in [
       constants.L7RULE_TYPE_SSL_CONN_HAS_CERT,
       constants.L7RULE_TYPE_SSL_VERIFY_RESULT,
       constants.L7RULE_TYPE_SSL_DN_FIELD]:
        return

    rule_type = None if l7rule.type == wtypes.Unset else l7rule.type
    req_key = None if l7rule.key == wtypes.Unset else l7rule.key
    req_value = None if l7rule.value == wtypes.Unset else l7rule.value
    compare_type = (None if l7rule.compare_type == wtypes.Unset else
                    l7rule.compare_type)
    msg = None
    if rule_type == constants.L7RULE_TYPE_SSL_CONN_HAS_CERT:
        # key and value are not allowed
        if req_key:
            # log error or raise
            msg = 'L7rule type {0} does not use the "key" field.'.format(
                rule_type)
        elif req_value.lower() != 'true':
            msg = 'L7rule value {0} is not a boolean True string.'.format(
                req_value)
        elif compare_type != constants.L7RULE_COMPARE_TYPE_EQUAL_TO:
            msg = 'L7rule type {0} only supports the {1} compare type.'.format(
                rule_type, constants.L7RULE_COMPARE_TYPE_EQUAL_TO)

    if rule_type == constants.L7RULE_TYPE_SSL_VERIFY_RESULT:
        if req_key:
            # log or raise req_key not used
            msg = 'L7rule type {0} does not use the "key" field.'.format(
                rule_type)
        elif not req_value.isdigit() or int(req_value) < 0:
            # log or raise req_value must be int
            msg = 'L7rule type {0} needs a int value, which is >= 0'.format(
                rule_type)
        elif compare_type != constants.L7RULE_COMPARE_TYPE_EQUAL_TO:
            msg = 'L7rule type {0} only supports the {1} compare type.'.format(
                rule_type, constants.L7RULE_COMPARE_TYPE_EQUAL_TO)

    if rule_type == constants.L7RULE_TYPE_SSL_DN_FIELD:
        dn_regex = re.compile(constants.DISTINGUISHED_NAME_FIELD_REGEX)
        if compare_type == constants.L7RULE_COMPARE_TYPE_REGEX:
            regex(l7rule.value)

        if not req_key or not req_value:
            # log or raise key and value must be specified.
            msg = 'L7rule type {0} needs to specify a key and a value.'.format(
                rule_type)
        # log or raise the key must be splited by '-'
        elif not dn_regex.match(req_key):
            msg = ('Invalid L7rule distinguished name field.')

    if msg:
        raise exceptions.InvalidL7Rule(msg=msg)


def sanitize_l7policy_api_args(l7policy, create=False):
    """Validate and make consistent L7Policy API arguments.

    This method is mainly meant to sanitize L7 Policy create and update
    API dictionaries, so that we strip 'None' values that don't apply for
    our particular update. This method does *not* verify that any
    redirect_pool_id exists in the database, but will raise an
    error if a redirect_url doesn't look like a URL.

    :param l7policy: The L7 Policy dictionary we are santizing / validating
    """
    if 'action' in l7policy.keys():
        if l7policy['action'] == constants.L7POLICY_ACTION_REJECT:
            l7policy.update({'redirect_url': None})
            l7policy.update({'redirect_pool_id': None})
            l7policy.pop('redirect_pool', None)
        elif l7policy['action'] == constants.L7POLICY_ACTION_REDIRECT_TO_URL:
            if not l7policy.get('redirect_url'):
                raise exceptions.InvalidL7PolicyArgs(
                    msg='redirect_url must not be None')
            l7policy.update({'redirect_pool_id': None})
            l7policy.pop('redirect_pool', None)
        elif l7policy['action'] == constants.L7POLICY_ACTION_REDIRECT_TO_POOL:
            if (not l7policy.get('redirect_pool_id') and
                    not l7policy.get('redirect_pool')):
                raise exceptions.InvalidL7PolicyArgs(
                    msg='redirect_pool_id or redirect_pool must not be None')
            l7policy.update({'redirect_url': None})
        elif l7policy['action'] == constants.L7POLICY_ACTION_REDIRECT_PREFIX:
            if not l7policy.get('redirect_prefix'):
                raise exceptions.InvalidL7PolicyArgs(
                    msg='redirect_prefix must not be None')
        else:
            raise exceptions.InvalidL7PolicyAction(
                action=l7policy['action'])
    if ((l7policy.get('redirect_pool_id') or l7policy.get('redirect_pool')) and
            (l7policy.get('redirect_url') or l7policy.get('redirect_prefix'))):
        raise exceptions.InvalidL7PolicyArgs(
            msg='Cannot specify redirect_pool_id and redirect_url or '
                'redirect_prefix at the same time')
    if l7policy.get('redirect_pool_id'):
        l7policy.update({
            'action': constants.L7POLICY_ACTION_REDIRECT_TO_POOL})
        l7policy.update({'redirect_url': None})
        l7policy.pop('redirect_pool', None)
        l7policy.update({'redirect_prefix': None})
        l7policy.update({'redirect_http_code': None})
    if l7policy.get('redirect_pool'):
        l7policy.update({
            'action': constants.L7POLICY_ACTION_REDIRECT_TO_POOL})
        l7policy.update({'redirect_url': None})
        l7policy.pop('redirect_pool_id', None)
        l7policy.update({'redirect_prefix': None})
        l7policy.update({'redirect_http_code': None})
    if l7policy.get('redirect_url'):
        url(l7policy['redirect_url'])
        l7policy.update({
            'action': constants.L7POLICY_ACTION_REDIRECT_TO_URL})
        l7policy.update({'redirect_pool_id': None})
        l7policy.update({'redirect_prefix': None})
        l7policy.pop('redirect_pool', None)
        if not l7policy.get('redirect_http_code'):
            l7policy.update({'redirect_http_code': 302})
    if l7policy.get('redirect_prefix'):
        url(l7policy['redirect_prefix'])
        l7policy.update({
            'action': constants.L7POLICY_ACTION_REDIRECT_PREFIX})
        l7policy.update({'redirect_pool_id': None})
        l7policy.update({'redirect_url': None})
        l7policy.pop('redirect_pool', None)
        if not l7policy.get('redirect_http_code'):
            l7policy.update({'redirect_http_code': 302})

    # If we are creating, we need an action at this point
    if create and 'action' not in l7policy.keys():
        raise exceptions.InvalidL7PolicyAction(action='None')

    # See if we have anything left after that...
    if not l7policy.keys():
        raise exceptions.InvalidL7PolicyArgs(msg='Invalid update options')
    return l7policy


def port_exists(port_id, context=None):
    """Raises an exception when a port does not exist."""
    network_driver = utils.get_network_driver()
    try:
        port = network_driver.get_port(port_id, context=context)
    except Exception as e:
        raise exceptions.InvalidSubresource(resource='Port', id=port_id) from e
    return port


def check_port_in_use(port):
    """Raise an exception when a port is used."""
    if port.device_id:
        raise exceptions.ValidationException(detail=_(
            "Port %(port_id)s is already used by device %(device_id)s ") %
            {'port_id': port.id, 'device_id': port.device_id})
    return False


def subnet_exists(subnet_id, context=None):
    """Raises an exception when a subnet does not exist."""
    network_driver = utils.get_network_driver()
    try:
        subnet = network_driver.get_subnet(subnet_id, context=context)
    except Exception as e:
        raise exceptions.InvalidSubresource(
            resource='Subnet', id=subnet_id) from e
    return subnet


def qos_policy_exists(qos_policy_id):
    network_driver = utils.get_network_driver()
    qos_extension_enabled(network_driver)
    try:
        qos_policy = network_driver.get_qos_policy(qos_policy_id)
    except Exception as e:
        raise exceptions.InvalidSubresource(
            resource='qos_policy', id=qos_policy_id) from e
    return qos_policy


def qos_extension_enabled(network_driver):
    if not network_driver.qos_enabled():
        raise exceptions.ValidationException(detail=_(
            "VIP QoS policy is not allowed in this deployment."))


def network_exists_optionally_contains_subnet(network_id, subnet_id=None,
                                              context=None):
    """Raises an exception when a network does not exist.

    If a subnet is provided, also validate the network contains that subnet.
    """
    network_driver = utils.get_network_driver()
    try:
        network = network_driver.get_network(network_id, context=context)
    except Exception as e:
        raise exceptions.InvalidSubresource(
            resource='Network', id=network_id) from e
    if subnet_id:
        if not network.subnets or subnet_id not in network.subnets:
            raise exceptions.InvalidSubresource(resource='Subnet',
                                                id=subnet_id)
    return network


def network_allowed_by_config(network_id, valid_networks=None):
    if CONF.networking.valid_vip_networks and not valid_networks:
        valid_networks = CONF.networking.valid_vip_networks
    if valid_networks:
        valid_networks = map(str.lower, valid_networks)
        if network_id.lower() not in valid_networks:
            raise exceptions.ValidationException(detail=_(
                'Supplied VIP network_id is not allowed by the configuration '
                'of this deployment.'))


def is_ip_member_of_cidr(address, cidr):
    if netaddr.IPAddress(address) in netaddr.IPNetwork(cidr):
        return True
    return False


def check_session_persistence(SP_dict):
    try:
        if SP_dict['cookie_name']:
            if SP_dict['type'] != constants.SESSION_PERSISTENCE_APP_COOKIE:
                raise exceptions.ValidationException(detail=_(
                    'Field "cookie_name" can only be specified with session '
                    'persistence of type "APP_COOKIE".'))
            bad_cookie_name = re.compile(r'[\x00-\x20\x22\x28-\x29\x2c\x2f'
                                         r'\x3a-\x40\x5b-\x5d\x7b\x7d\x7f]+')
            valid_chars = re.compile(r'[\x00-\xff]+')
            if (bad_cookie_name.search(SP_dict['cookie_name']) or
                    not valid_chars.search(SP_dict['cookie_name'])):
                raise exceptions.ValidationException(detail=_(
                    'Supplied "cookie_name" is invalid.'))
        if (SP_dict['type'] == constants.SESSION_PERSISTENCE_APP_COOKIE and
                not SP_dict['cookie_name']):
            raise exceptions.ValidationException(detail=_(
                'Field "cookie_name" must be specified when using the '
                '"APP_COOKIE" session persistence type.'))
    except exceptions.ValidationException:
        raise
    except Exception as e:
        raise exceptions.ValidationException(detail=_(
            'Invalid session_persistence provided.')) from e


def ip_not_reserved(ip_address):
    ip_address = (
        ipaddress.ip_address(ip_address).exploded.upper())
    if ip_address in CONF.networking.reserved_ips:
        raise exceptions.InvalidOption(value=ip_address,
                                       option='member address')


def check_cipher_prohibit_list(cipherstring):
    ciphers = cipherstring.split(':')
    prohibit_list = CONF.api_settings.tls_cipher_prohibit_list.split(':')
    rejected = []
    for cipher in ciphers:
        if cipher in prohibit_list:
            rejected.append(cipher)
    return rejected


def check_default_ciphers_prohibit_list_conflict():
    listener_rejected = check_cipher_prohibit_list(
        CONF.api_settings.default_listener_ciphers)
    if listener_rejected:
        raise exceptions.ValidationException(
            detail=_('Default listener ciphers conflict with prohibit list. '
                     'Conflicting ciphers: ' + ', '.join(listener_rejected)))

    pool_rejected = check_cipher_prohibit_list(
        CONF.api_settings.default_pool_ciphers)
    if pool_rejected:
        raise exceptions.ValidationException(
            detail=_('Default pool ciphers conflict with prohibit list. '
                     'Conflicting ciphers: ' + ', '.join(pool_rejected)))


def check_tls_version_list(versions):
    if versions == []:
        raise exceptions.ValidationException(
            detail=_('Empty TLS version list. Either specify at least one TLS '
                     'version or remove this parameter to use the default.'))

    # Unset action
    if versions is None:
        return

    invalid_versions = [v for v in versions
                        if v not in constants.TLS_ALL_VERSIONS]
    if invalid_versions:
        raise exceptions.ValidationException(
            detail=_('Invalid TLS versions: ' + ', '.join(invalid_versions)))


def check_tls_version_min(versions, message=None):
    """Checks a TLS version string against the configured minimum."""

    if not CONF.api_settings.minimum_tls_version:
        return

    if not message:
        message = _("Requested TLS versions are less than the minimum: ")

    min_ver_index = constants.TLS_ALL_VERSIONS.index(
        CONF.api_settings.minimum_tls_version)

    rejected = []
    for ver in versions:
        if constants.TLS_ALL_VERSIONS.index(ver) < min_ver_index:
            rejected.append(ver)
    if rejected:
        raise exceptions.ValidationException(detail=(
            message + ', '.join(rejected) + " < " +
            CONF.api_settings.minimum_tls_version))


def check_default_tls_versions_min_conflict():
    if not CONF.api_settings.minimum_tls_version:
        return

    listener_message = _("Default listener TLS versions are less than the "
                         "minimum: ")
    pool_message = _("Default pool TLS versions are less than the minimum: ")

    check_tls_version_min(CONF.api_settings.default_listener_tls_versions,
                          message=listener_message)

    check_tls_version_min(CONF.api_settings.default_pool_tls_versions,
                          message=pool_message)


def check_alpn_protocols(protocols):
    if protocols == []:
        raise exceptions.ValidationException(
            detail=_('Empty ALPN protocol list. Either specify at least one '
                     'ALPN protocol or remove this parameter to use the '
                     'default.'))

    # Unset action
    if protocols is None:
        return

    invalid_protocols = [p for p in protocols
                         if p not in constants.SUPPORTED_ALPN_PROTOCOLS]
    if invalid_protocols:
        raise exceptions.ValidationException(
            detail=_('Invalid ALPN protocol: ' + ', '.join(invalid_protocols)))



# Copyright 2021 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Add prometheus listener protocol

Revision ID: 6ac558d7fc21
Revises: b8bd389cbae7
Create Date: 2021-10-01 20:06:46.813842

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy import sql

# revision identifiers, used by Alembic.
revision = '6ac558d7fc21'
down_revision = 'b8bd389cbae7'

new_protocol = 'PROMETHEUS'


def upgrade():
    insert_table = sql.table(
        u'protocol',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': new_protocol}
        ]
    )



#    Copyright 2018 Rackspace, US Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""add_flavor_id_to_lb

Revision ID: 211982b05afc
Revises: b9c703669314
Create Date: 2018-11-30 14:57:28.559884

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '211982b05afc'
down_revision = 'b9c703669314'


def upgrade():
    op.add_column('load_balancer',
                  sa.Column('flavor_id', sa.String(36), nullable=True))



#    Copyright 2016 Intel Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture
from oslo_utils import uuidutils

from octavia.api.common import pagination
from octavia.common import exceptions
from octavia.db import models
from octavia.tests.unit import base

DEFAULT_SORTS = [('created_at', 'asc'), ('id', 'asc')]


class TestPaginationHelper(base.TestCase):

    @mock.patch('octavia.api.common.pagination.request')
    def test_no_params(self, request_mock):
        params = {}
        helper = pagination.PaginationHelper(params)
        query_mock = mock.MagicMock()

        helper.apply(query_mock, models.LoadBalancer)
        self.assertEqual(DEFAULT_SORTS, helper.sort_keys)
        self.assertIsNone(helper.marker)
        self.assertEqual(1000, helper.limit)
        query_mock.order_by().order_by().limit.assert_called_with(
            1000)

    def test_sort_empty(self):
        sort_params = ""
        params = {'sort': sort_params}
        act_params = pagination.PaginationHelper(
            params).sort_keys
        self.assertEqual([], act_params)

    def test_sort_none(self):
        sort_params = None
        params = {'sort': sort_params}
        act_params = pagination.PaginationHelper(
            params).sort_keys
        self.assertEqual([], act_params)

    def test_sort_key_dir(self):
        sort_keys = "key1,key2,key3"
        sort_dirs = "asc,desc"
        ref_sort_keys = [('key1', 'asc'), ('key2', 'desc'), ('key3', 'asc')]
        params = {'sort_key': sort_keys, 'sort_dir': sort_dirs}
        helper = pagination.PaginationHelper(params)
        self.assertEqual(ref_sort_keys, helper.sort_keys)

    def test_invalid_sorts(self):
        sort_params = "shoud_fail_exception:cause:of:this"
        params = {'sort': sort_params}
        self.assertRaises(exceptions.InvalidSortKey,
                          pagination.PaginationHelper,
                          params)

        sort_params = "ke1:asc,key2:InvalidDir,key3"
        params = {'sort': sort_params}
        self.assertRaises(exceptions.InvalidSortDirection,
                          pagination.PaginationHelper,
                          params)

    def test_marker(self):
        marker = 'random_uuid'
        params = {'marker': marker}
        helper = pagination.PaginationHelper(params)

        self.assertEqual(marker, helper.marker)

    @mock.patch('octavia.api.common.pagination.request')
    def test_limit(self, request_mock):
        limit = 100
        params = {'limit': limit}
        helper = pagination.PaginationHelper(params)
        query_mock = mock.MagicMock()

        helper.apply(query_mock, models.LoadBalancer)
        query_mock.order_by().order_by().limit.assert_called_with(
            limit)

    @mock.patch('octavia.api.common.pagination.request')
    def test_filter_correct_params(self, request_mock):
        params = {'id': 'fake_id'}
        helper = pagination.PaginationHelper(params)
        query_mock = mock.MagicMock()

        helper.apply(query_mock, models.LoadBalancer)
        self.assertEqual(params, helper.filters)

    @mock.patch('octavia.api.common.pagination.request')
    def test_filter_mismatched_params(self, request_mock):
        params = {
            'id': 'fake_id',
            'fields': 'field',
            'limit': '10',
            'sort': None,
        }

        filters = {'id': 'fake_id'}

        helper = pagination.PaginationHelper(params)
        query_mock = mock.MagicMock()

        helper.apply(query_mock, models.LoadBalancer)
        self.assertEqual(filters, helper.filters)
        helper.apply(query_mock, models.LoadBalancer,
                     enforce_valid_params=True)
        self.assertEqual(filters, helper.filters)

    @mock.patch('octavia.api.common.pagination.request')
    def test_filter_with_invalid_params(self, request_mock):
        params = {'id': 'fake_id', 'no_such_param': 'id'}
        filters = {'id': 'fake_id'}
        helper = pagination.PaginationHelper(params)
        query_mock = mock.MagicMock()

        helper.apply(query_mock, models.LoadBalancer,
                     # silently ignore invalid parameter
                     enforce_valid_params=False)
        self.assertEqual(filters, helper.filters)

        self.assertRaises(
            exceptions.InvalidFilterArgument,
            pagination.PaginationHelper.apply,
            helper,
            query_mock,
            models.Amphora,
        )

    @mock.patch('octavia.api.common.pagination.request')
    def test_duplicate_argument(self, request_mock):
        params = {'loadbalacer_id': 'id1', 'load_balacer_id': 'id2'}
        query_mock = mock.MagicMock()
        helper = pagination.PaginationHelper(params)

        self.assertRaises(
            exceptions.InvalidFilterArgument,
            pagination.PaginationHelper.apply,
            helper,
            query_mock,
            models.Amphora,
        )

    @mock.patch('octavia.api.common.pagination.request')
    def test_fields_not_passed(self, request_mock):
        params = {'fields': 'id'}
        helper = pagination.PaginationHelper(params)
        query_mock = mock.MagicMock()

        helper.apply(query_mock, models.LoadBalancer)
        self.assertEqual({}, helper.filters)

    @mock.patch('octavia.api.common.pagination.request')
    def test_make_links_next(self, request_mock):
        request_mock.path = "/lbaas/v2/pools/1/members"
        request_mock.path_url = "http://localhost" + request_mock.path
        member1 = models.Member()
        member1.id = uuidutils.generate_uuid()
        model_list = [member1]

        params = {'limit': 1}
        helper = pagination.PaginationHelper(params)
        links = helper._make_links(model_list)
        self.assertEqual(links[0].rel, "next")
        self.assertEqual(
            links[0].href,
            "{path_url}?limit={limit}&marker={marker}".format(
                path_url=request_mock.path_url,
                limit=params['limit'],
                marker=member1.id
            ))

    @mock.patch('octavia.api.common.pagination.request')
    def test_make_links_prev(self, request_mock):
        request_mock.path = "/lbaas/v2/pools/1/members"
        request_mock.path_url = "http://localhost" + request_mock.path
        member1 = models.Member()
        member1.id = uuidutils.generate_uuid()
        model_list = [member1]

        params = {'limit': 1, 'marker': member1.id}
        helper = pagination.PaginationHelper(params)
        links = helper._make_links(model_list)
        self.assertEqual(links[0].rel, "previous")
        self.assertEqual(
            links[1].href,
            "{path_url}?limit={limit}&marker={marker}".format(
                path_url=request_mock.path_url,
                limit=params['limit'],
                marker=member1.id))
        self.assertEqual(links[1].rel, "next")
        self.assertEqual(
            links[1].href,
            "{path_url}?limit={limit}&marker={marker}".format(
                path_url=request_mock.path_url,
                limit=params['limit'],
                marker=member1.id))

    @mock.patch('octavia.api.common.pagination.request')
    def test_make_links_with_configured_url(self, request_mock):
        request_mock.path = "/lbaas/v2/pools/1/members"
        request_mock.path_url = "http://localhost" + request_mock.path
        api_base_uri = "https://127.0.0.1"
        conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        conf.config(group='api_settings', api_base_uri=api_base_uri)
        member1 = models.Member()
        member1.id = uuidutils.generate_uuid()
        model_list = [member1]

        params = {'limit': 1, 'marker': member1.id}
        helper = pagination.PaginationHelper(params)
        links = helper._make_links(model_list)
        self.assertEqual(links[0].rel, "previous")
        self.assertEqual(
            links[1].href,
            "{base_uri}{path}?limit={limit}&marker={marker}".format(
                base_uri=api_base_uri,
                path=request_mock.path,
                limit=params['limit'],
                marker=member1.id
            ))
        self.assertEqual(links[1].rel, "next")
        self.assertEqual(
            links[1].href,
            "{base_uri}{path}?limit={limit}&marker={marker}".format(
                base_uri=api_base_uri,
                path=request_mock.path,
                limit=params['limit'],
                marker=member1.id))



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture
from oslo_utils import uuidutils
from taskflow.patterns import linear_flow as flow

from octavia.common import constants
from octavia.common import data_models
from octavia.controller.worker.v2.flows import amphora_flows
import octavia.tests.unit.base as base

AUTH_VERSION = '2'


# NOTE: We patch the get_network_driver for all the calls so we don't
# inadvertently make real calls.
@mock.patch('octavia.common.utils.get_network_driver')
class TestAmphoraFlows(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        self.conf.config(
            group="controller_worker",
            amphora_driver='amphora_haproxy_rest_driver')
        self.conf.config(group="nova", enable_anti_affinity=False)
        self.AmpFlow = amphora_flows.AmphoraFlows()
        self.amp1 = data_models.Amphora(id=1)
        self.amp2 = data_models.Amphora(id=2)
        self.amp3 = data_models.Amphora(id=3, status=constants.DELETED)
        self.amp4 = data_models.Amphora(id=uuidutils.generate_uuid())
        self.lb = data_models.LoadBalancer(
            id=4, amphorae=[self.amp1, self.amp2, self.amp3])

    def test_get_create_amphora_flow(self, mock_get_net_driver):

        amp_flow = self.AmpFlow.get_create_amphora_flow()

        self.assertIsInstance(amp_flow, flow.Flow)

        self.assertIn(constants.AMPHORA, amp_flow.provides)
        self.assertIn(constants.AMPHORA_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_OBJ, amp_flow.provides)
        self.assertIn(constants.SERVER_PEM, amp_flow.provides)

        self.assertIn(constants.BUILD_TYPE_PRIORITY, amp_flow.requires)
        self.assertIn(constants.FLAVOR, amp_flow.requires)
        self.assertIn(constants.AVAILABILITY_ZONE, amp_flow.requires)

        self.assertEqual(5, len(amp_flow.provides))
        self.assertEqual(4, len(amp_flow.requires))

    def test_get_create_amphora_flow_cert(self, mock_get_net_driver):
        self.AmpFlow = amphora_flows.AmphoraFlows()

        amp_flow = self.AmpFlow.get_create_amphora_flow()

        self.assertIsInstance(amp_flow, flow.Flow)

        self.assertIn(constants.AMPHORA, amp_flow.provides)
        self.assertIn(constants.AMPHORA_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_OBJ, amp_flow.provides)
        self.assertIn(constants.SERVER_PEM, amp_flow.provides)

        self.assertIn(constants.BUILD_TYPE_PRIORITY, amp_flow.requires)
        self.assertIn(constants.FLAVOR, amp_flow.requires)
        self.assertIn(constants.AVAILABILITY_ZONE, amp_flow.requires)

        self.assertEqual(5, len(amp_flow.provides))
        self.assertEqual(4, len(amp_flow.requires))

    def test_get_amphora_for_lb_flow(self, mock_get_net_driver):

        amp_flow = self.AmpFlow.get_amphora_for_lb_subflow(
            'SOMEPREFIX', constants.ROLE_STANDALONE)

        self.assertIsInstance(amp_flow, flow.Flow)

        self.assertIn(constants.FLAVOR, amp_flow.requires)
        self.assertIn(constants.AVAILABILITY_ZONE, amp_flow.requires)
        self.assertIn(constants.BUILD_TYPE_PRIORITY, amp_flow.requires)
        self.assertIn(constants.LOADBALANCER_ID, amp_flow.requires)

        self.assertIn(constants.AMPHORA, amp_flow.provides)
        self.assertIn(constants.AMPHORA_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_OBJ, amp_flow.provides)
        self.assertIn(constants.SERVER_PEM, amp_flow.provides)

        self.assertEqual(5, len(amp_flow.provides))
        self.assertEqual(5, len(amp_flow.requires))

    def test_get_cert_create_amphora_for_lb_flow(self, mock_get_net_driver):

        self.AmpFlow = amphora_flows.AmphoraFlows()

        amp_flow = self.AmpFlow.get_amphora_for_lb_subflow(
            'SOMEPREFIX', constants.ROLE_STANDALONE)

        self.assertIsInstance(amp_flow, flow.Flow)

        self.assertIn(constants.FLAVOR, amp_flow.requires)
        self.assertIn(constants.AVAILABILITY_ZONE, amp_flow.requires)
        self.assertIn(constants.BUILD_TYPE_PRIORITY, amp_flow.requires)
        self.assertIn(constants.LOADBALANCER_ID, amp_flow.requires)

        self.assertIn(constants.AMPHORA, amp_flow.provides)
        self.assertIn(constants.AMPHORA_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_OBJ, amp_flow.provides)
        self.assertIn(constants.SERVER_PEM, amp_flow.provides)

        self.assertEqual(5, len(amp_flow.provides))
        self.assertEqual(5, len(amp_flow.requires))

    def test_get_cert_master_create_amphora_for_lb_flow(
            self, mock_get_net_driver):

        self.AmpFlow = amphora_flows.AmphoraFlows()

        amp_flow = self.AmpFlow.get_amphora_for_lb_subflow(
            'SOMEPREFIX', constants.ROLE_MASTER)

        self.assertIsInstance(amp_flow, flow.Flow)

        self.assertIn(constants.FLAVOR, amp_flow.requires)
        self.assertIn(constants.AVAILABILITY_ZONE, amp_flow.requires)
        self.assertIn(constants.BUILD_TYPE_PRIORITY, amp_flow.requires)
        self.assertIn(constants.LOADBALANCER_ID, amp_flow.requires)

        self.assertIn(constants.AMPHORA, amp_flow.provides)
        self.assertIn(constants.AMPHORA_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_OBJ, amp_flow.provides)
        self.assertIn(constants.SERVER_PEM, amp_flow.provides)

        self.assertEqual(5, len(amp_flow.provides))
        self.assertEqual(5, len(amp_flow.requires))

    def test_get_cert_master_rest_anti_affinity_create_amphora_for_lb_flow(
            self, mock_get_net_driver):

        self.conf.config(group="nova", enable_anti_affinity=True)

        self.AmpFlow = amphora_flows.AmphoraFlows()
        amp_flow = self.AmpFlow.get_amphora_for_lb_subflow(
            'SOMEPREFIX', constants.ROLE_MASTER)

        self.assertIsInstance(amp_flow, flow.Flow)

        self.assertIn(constants.FLAVOR, amp_flow.requires)
        self.assertIn(constants.AVAILABILITY_ZONE, amp_flow.requires)
        self.assertIn(constants.BUILD_TYPE_PRIORITY, amp_flow.requires)
        self.assertIn(constants.LOADBALANCER_ID, amp_flow.requires)
        self.assertIn(constants.SERVER_GROUP_ID, amp_flow.requires)

        self.assertIn(constants.AMPHORA, amp_flow.provides)
        self.assertIn(constants.AMPHORA_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_OBJ, amp_flow.provides)
        self.assertIn(constants.SERVER_PEM, amp_flow.provides)

        self.assertEqual(5, len(amp_flow.provides))
        self.assertEqual(5, len(amp_flow.requires))
        self.conf.config(group="nova", enable_anti_affinity=False)

    def test_get_cert_backup_create_amphora_for_lb_flow(
            self, mock_get_net_driver):
        self.AmpFlow = amphora_flows.AmphoraFlows()

        amp_flow = self.AmpFlow.get_amphora_for_lb_subflow(
            'SOMEPREFIX', constants.ROLE_BACKUP)

        self.assertIsInstance(amp_flow, flow.Flow)

        self.assertIn(constants.FLAVOR, amp_flow.requires)
        self.assertIn(constants.AVAILABILITY_ZONE, amp_flow.requires)
        self.assertIn(constants.BUILD_TYPE_PRIORITY, amp_flow.requires)
        self.assertIn(constants.LOADBALANCER_ID, amp_flow.requires)

        self.assertIn(constants.AMPHORA, amp_flow.provides)
        self.assertIn(constants.AMPHORA_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_OBJ, amp_flow.provides)
        self.assertIn(constants.SERVER_PEM, amp_flow.provides)

        self.assertEqual(5, len(amp_flow.provides))
        self.assertEqual(5, len(amp_flow.requires))

    def test_get_cert_bogus_create_amphora_for_lb_flow(
            self, mock_get_net_driver):
        self.AmpFlow = amphora_flows.AmphoraFlows()

        amp_flow = self.AmpFlow.get_amphora_for_lb_subflow(
            'SOMEPREFIX', 'BOGUS_ROLE')

        self.assertIsInstance(amp_flow, flow.Flow)

        self.assertIn(constants.FLAVOR, amp_flow.requires)
        self.assertIn(constants.AVAILABILITY_ZONE, amp_flow.requires)
        self.assertIn(constants.BUILD_TYPE_PRIORITY, amp_flow.requires)
        self.assertIn(constants.LOADBALANCER_ID, amp_flow.requires)

        self.assertIn(constants.AMPHORA, amp_flow.provides)
        self.assertIn(constants.AMPHORA_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_OBJ, amp_flow.provides)
        self.assertIn(constants.SERVER_PEM, amp_flow.provides)

        self.assertEqual(5, len(amp_flow.provides))
        self.assertEqual(5, len(amp_flow.requires))

    def test_get_cert_backup_rest_anti_affinity_create_amphora_for_lb_flow(
            self, mock_get_net_driver):
        self.conf.config(group="nova", enable_anti_affinity=True)

        self.AmpFlow = amphora_flows.AmphoraFlows()
        amp_flow = self.AmpFlow.get_amphora_for_lb_subflow(
            'SOMEPREFIX', constants.ROLE_BACKUP)

        self.assertIsInstance(amp_flow, flow.Flow)

        self.assertIn(constants.FLAVOR, amp_flow.requires)
        self.assertIn(constants.AVAILABILITY_ZONE, amp_flow.requires)
        self.assertIn(constants.BUILD_TYPE_PRIORITY, amp_flow.requires)
        self.assertIn(constants.LOADBALANCER_ID, amp_flow.requires)
        self.assertIn(constants.SERVER_GROUP_ID, amp_flow.requires)

        self.assertIn(constants.AMPHORA_ID, amp_flow.provides)
        self.assertIn(constants.SERVER_GROUP_ID, amp_flow.requires)
        self.assertIn(constants.COMPUTE_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_OBJ, amp_flow.provides)
        self.assertIn(constants.SERVER_PEM, amp_flow.provides)

        self.assertEqual(5, len(amp_flow.provides))
        self.assertEqual(5, len(amp_flow.requires))
        self.conf.config(group="nova", enable_anti_affinity=False)

    def test_get_delete_amphora_flow(self, mock_get_net_driver):

        amp_flow = self.AmpFlow.get_delete_amphora_flow(self.amp4.to_dict())

        self.assertIsInstance(amp_flow, flow.Flow)

        # This flow injects the required data at flow compile time.

        self.assertEqual(0, len(amp_flow.provides))
        self.assertEqual(0, len(amp_flow.requires))

    def test_get_failover_flow_act_stdby(self, mock_get_net_driver):
        failed_amphora = data_models.Amphora(
            id=uuidutils.generate_uuid(), role=constants.ROLE_MASTER,
            load_balancer_id=uuidutils.generate_uuid()).to_dict()

        amp_flow = self.AmpFlow.get_failover_amphora_flow(
            failed_amphora, 2)

        self.assertIsInstance(amp_flow, flow.Flow)

        self.assertIn(constants.AVAILABILITY_ZONE, amp_flow.requires)
        self.assertIn(constants.BUILD_TYPE_PRIORITY, amp_flow.requires)
        self.assertIn(constants.FLAVOR, amp_flow.requires)
        self.assertIn(constants.LOADBALANCER, amp_flow.requires)
        self.assertIn(constants.LOADBALANCER_ID, amp_flow.requires)
        self.assertIn(constants.VIP, amp_flow.requires)

        self.assertIn(constants.ADDED_PORTS, amp_flow.provides)
        self.assertIn(constants.AMP_VRRP_INT, amp_flow.provides)
        self.assertIn(constants.AMPHORA, amp_flow.provides)
        self.assertIn(constants.AMPHORA_ID, amp_flow.provides)
        self.assertIn(constants.AMPHORAE, amp_flow.provides)
        self.assertIn(constants.AMPHORAE_NETWORK_CONFIG, amp_flow.provides)
        self.assertIn(constants.BASE_PORT, amp_flow.provides)
        self.assertIn(constants.COMPUTE_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_OBJ, amp_flow.provides)
        self.assertIn(constants.DELTA, amp_flow.provides)
        self.assertIn(constants.LOADBALANCER, amp_flow.provides)
        self.assertIn(constants.SERVER_PEM, amp_flow.provides)
        self.assertIn(constants.VIP_SG_ID, amp_flow.provides)

        self.assertEqual(7, len(amp_flow.requires))
        self.assertEqual(13, len(amp_flow.provides))

    def test_get_failover_flow_standalone(self, mock_get_net_driver):
        failed_amphora = data_models.Amphora(
            id=uuidutils.generate_uuid(), role=constants.ROLE_STANDALONE,
            load_balancer_id=uuidutils.generate_uuid(),
            vrrp_ip='2001:3b8::32').to_dict()

        amp_flow = self.AmpFlow.get_failover_amphora_flow(
            failed_amphora, 1)

        self.assertIsInstance(amp_flow, flow.Flow)

        self.assertIn(constants.AVAILABILITY_ZONE, amp_flow.requires)
        self.assertIn(constants.BUILD_TYPE_PRIORITY, amp_flow.requires)
        self.assertIn(constants.FLAVOR, amp_flow.requires)
        self.assertIn(constants.LOADBALANCER, amp_flow.requires)
        self.assertIn(constants.LOADBALANCER_ID, amp_flow.requires)
        self.assertIn(constants.VIP, amp_flow.requires)

        self.assertIn(constants.ADDED_PORTS, amp_flow.provides)
        self.assertIn(constants.AMPHORA, amp_flow.provides)
        self.assertIn(constants.AMPHORA_ID, amp_flow.provides)
        self.assertIn(constants.AMPHORAE, amp_flow.provides)
        self.assertIn(constants.AMPHORAE_NETWORK_CONFIG, amp_flow.provides)
        self.assertIn(constants.BASE_PORT, amp_flow.provides)
        self.assertIn(constants.COMPUTE_ID, amp_flow.provides)
        self.assertIn(constants.COMPUTE_OBJ, amp_flow.provides)
        self.assertIn(constants.DELTA, amp_flow.provides)
        self.assertIn(constants.LOADBALANCER, amp_flow.provides)
        self.assertIn(constants.SERVER_PEM, amp_flow.provides)
        self.assertIn(constants.VIP_SG_ID, amp_flow.provides)

        self.assertEqual(7, len(amp_flow.requires))
        self.assertEqual(12, len(amp_flow.provides))

    def test_get_failover_flow_bogus_role(self, mock_get_net_driver):
        failed_amphora = data_models.Amphora(id=uuidutils.generate_uuid(),
                                             role='bogus').to_dict()

        amp_flow = self.AmpFlow.get_failover_amphora_flow(
            failed_amphora, 1)

        self.assertIsInstance(amp_flow, flow.Flow)

        self.assertIn(constants.LOADBALANCER_ID, amp_flow.requires)
        self.assertIn(constants.LOADBALANCER, amp_flow.requires)

        self.assertIn(constants.VIP_SG_ID, amp_flow.provides)

        self.assertEqual(2, len(amp_flow.requires))
        self.assertEqual(1, len(amp_flow.provides))

    def test_cert_rotate_amphora_flow(self, mock_get_net_driver):
        self.AmpFlow = amphora_flows.AmphoraFlows()

        amp_rotate_flow = self.AmpFlow.cert_rotate_amphora_flow()
        self.assertIsInstance(amp_rotate_flow, flow.Flow)

        self.assertIn(constants.SERVER_PEM, amp_rotate_flow.provides)
        self.assertIn(constants.AMPHORA, amp_rotate_flow.requires)

        self.assertEqual(1, len(amp_rotate_flow.provides))
        self.assertEqual(2, len(amp_rotate_flow.requires))

    def test_get_vrrp_subflow(self, mock_get_net_driver):
        vrrp_subflow = self.AmpFlow.get_vrrp_subflow('123')

        self.assertIsInstance(vrrp_subflow, flow.Flow)

        self.assertIn(constants.AMPHORAE_NETWORK_CONFIG, vrrp_subflow.provides)
        self.assertIn(constants.AMP_VRRP_INT, vrrp_subflow.provides)

        self.assertIn(constants.LOADBALANCER_ID, vrrp_subflow.requires)
        self.assertIn(constants.AMPHORAE, vrrp_subflow.requires)

        self.assertEqual(2, len(vrrp_subflow.provides))
        self.assertEqual(2, len(vrrp_subflow.requires))

    def test_get_vrrp_subflow_dont_create_vrrp_group(
            self, mock_get_net_driver):
        vrrp_subflow = self.AmpFlow.get_vrrp_subflow('123',
                                                     create_vrrp_group=False)

        self.assertIsInstance(vrrp_subflow, flow.Flow)

        self.assertIn(constants.AMPHORAE_NETWORK_CONFIG, vrrp_subflow.provides)
        self.assertIn(constants.AMP_VRRP_INT, vrrp_subflow.provides)

        self.assertIn(constants.LOADBALANCER_ID, vrrp_subflow.requires)
        self.assertIn(constants.AMPHORAE, vrrp_subflow.requires)

        self.assertEqual(2, len(vrrp_subflow.provides))
        self.assertEqual(2, len(vrrp_subflow.requires))

    def test_update_amphora_config_flow(self, mock_get_net_driver):

        amp_flow = self.AmpFlow.update_amphora_config_flow()

        self.assertIsInstance(amp_flow, flow.Flow)

        self.assertIn(constants.AMPHORA, amp_flow.requires)
        self.assertIn(constants.FLAVOR, amp_flow.requires)

        self.assertEqual(2, len(amp_flow.requires))
        self.assertEqual(0, len(amp_flow.provides))

    def test__retry_flow(self, mock_get_net_driver):
        amp_flow = self.AmpFlow._retry_flow('test_flow')

        self.assertIsInstance(amp_flow, flow.Flow)

        self.assertIn(constants.AMPHORA, amp_flow.requires)

        self.assertEqual(1, len(amp_flow.requires))
        self.assertEqual(0, len(amp_flow.provides))



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

from taskflow.patterns import linear_flow as flow

from octavia.common import constants
from octavia.controller.worker.v1.flows import health_monitor_flows
import octavia.tests.unit.base as base


class TestHealthMonitorFlows(base.TestCase):

    def setUp(self):
        self.HealthMonitorFlow = health_monitor_flows.HealthMonitorFlows()

        super().setUp()

    def test_get_create_health_monitor_flow(self):

        health_mon_flow = (self.HealthMonitorFlow.
                           get_create_health_monitor_flow())

        self.assertIsInstance(health_mon_flow, flow.Flow)

        self.assertIn(constants.LISTENERS, health_mon_flow.requires)
        self.assertIn(constants.LOADBALANCER, health_mon_flow.requires)
        self.assertIn(constants.POOL, health_mon_flow.requires)

        self.assertEqual(4, len(health_mon_flow.requires))
        self.assertEqual(0, len(health_mon_flow.provides))

    def test_get_delete_health_monitor_flow(self):

        health_mon_flow = (self.HealthMonitorFlow.
                           get_delete_health_monitor_flow())

        self.assertIsInstance(health_mon_flow, flow.Flow)

        self.assertIn(constants.HEALTH_MON, health_mon_flow.requires)
        self.assertIn(constants.LISTENERS, health_mon_flow.requires)
        self.assertIn(constants.LOADBALANCER, health_mon_flow.requires)
        self.assertIn(constants.POOL, health_mon_flow.requires)

        self.assertEqual(4, len(health_mon_flow.requires))
        self.assertEqual(0, len(health_mon_flow.provides))

    def test_get_update_health_monitor_flow(self):

        health_mon_flow = (self.HealthMonitorFlow.
                           get_update_health_monitor_flow())

        self.assertIsInstance(health_mon_flow, flow.Flow)

        self.assertIn(constants.LISTENERS, health_mon_flow.requires)
        self.assertIn(constants.LOADBALANCER, health_mon_flow.requires)
        self.assertIn(constants.HEALTH_MON, health_mon_flow.requires)
        self.assertIn(constants.UPDATE_DICT, health_mon_flow.requires)

        self.assertEqual(5, len(health_mon_flow.requires))
        self.assertEqual(0, len(health_mon_flow.provides))



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import ctypes
import os


class NetworkNamespace(object):
    """A network namespace context manager.

    Runs wrapped code inside the specified network namespace.

    :param netns: The network namespace name to enter.
    """
    # from linux/sched.h - We want to enter a network namespace
    CLONE_NEWNET = 0x40000000

    @staticmethod
    def _error_handler(result, func, arguments):
        if result == -1:
            errno = ctypes.get_errno()
            raise OSError(errno, os.strerror(errno))

    def __init__(self, netns):
        self.current_netns = '/proc/{pid}/ns/net'.format(pid=os.getpid())
        self.target_netns = '/var/run/netns/{netns}'.format(netns=netns)
        # reference: man setns(2)
        self.set_netns = ctypes.CDLL('libc.so.6', use_errno=True).setns
        self.set_netns.errcheck = self._error_handler

    def __enter__(self):
        # Save the current network namespace
        self.current_netns_fd = open(self.current_netns, encoding='utf-8')
        with open(self.target_netns, encoding='utf-8') as fd:
            self.set_netns(fd.fileno(), self.CLONE_NEWNET)

    def __exit__(self, *args):
        # Return to the previous network namespace
        self.set_netns(self.current_netns_fd.fileno(), self.CLONE_NEWNET)
        self.current_netns_fd.close()



#    Copyright 2019 Verizon Media
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from wsme import types as wtypes

from octavia.api.common import types


class BaseAvailabilityZoneType(types.BaseType):
    _type_to_model_map = {}
    _child_map = {}


class AvailabilityZoneResponse(BaseAvailabilityZoneType):
    """Defines which attributes are to be shown on any response."""
    name = wtypes.wsattr(wtypes.StringType())
    description = wtypes.wsattr(wtypes.StringType())
    enabled = wtypes.wsattr(bool)
    availability_zone_profile_id = wtypes.wsattr(wtypes.StringType())

    @classmethod
    def from_data_model(cls, data_model, children=False):
        availability_zone = super(
            AvailabilityZoneResponse, cls).from_data_model(
            data_model, children=children)
        return availability_zone


class AvailabilityZoneRootResponse(types.BaseType):
    availability_zone = wtypes.wsattr(AvailabilityZoneResponse)


class AvailabilityZonesRootResponse(types.BaseType):
    availability_zones = wtypes.wsattr([AvailabilityZoneResponse])
    availability_zones_links = wtypes.wsattr([types.PageType])


class AvailabilityZonePOST(BaseAvailabilityZoneType):
    """Defines mandatory and optional attributes of a POST request."""
    name = wtypes.wsattr(wtypes.StringType(max_length=255), mandatory=True)
    description = wtypes.wsattr(wtypes.StringType(max_length=255))
    enabled = wtypes.wsattr(bool, default=True)
    availability_zone_profile_id = wtypes.wsattr(wtypes.UuidType(),
                                                 mandatory=True)


class AvailabilityZoneRootPOST(types.BaseType):
    availability_zone = wtypes.wsattr(AvailabilityZonePOST)


class AvailabilityZonePUT(BaseAvailabilityZoneType):
    """Defines the attributes of a PUT request."""
    description = wtypes.wsattr(wtypes.StringType(max_length=255))
    enabled = wtypes.wsattr(bool)


class AvailabilityZoneRootPUT(types.BaseType):
    availability_zone = wtypes.wsattr(AvailabilityZonePUT)



# Copyright (c) 2014 Rackspace US, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Legacy cert manager implementation for Barbican (container+secrets)
"""
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from stevedore import driver as stevedore_driver

from octavia.certificates.common import barbican as barbican_common
from octavia.certificates.manager import cert_mgr
from octavia.common.tls_utils import cert_parser

LOG = logging.getLogger(__name__)


class BarbicanCertManager(cert_mgr.CertManager):
    """Certificate Manager that wraps the Barbican client API."""

    def __init__(self, auth=None):
        super().__init__()
        if auth:
            self.auth = auth
        else:
            self.auth = stevedore_driver.DriverManager(
                namespace='octavia.barbican_auth',
                name=cfg.CONF.certificates.barbican_auth,
                invoke_on_load=True,
            ).driver

    def store_cert(self, context, certificate, private_key, intermediates=None,
                   private_key_passphrase=None, expiration=None, name=None):
        """Stores a certificate in the certificate manager.

        :param context: Oslo context of the request
        :param certificate: PEM encoded TLS certificate
        :param private_key: private key for the supplied certificate
        :param intermediates: ordered and concatenated intermediate certs
        :param private_key_passphrase: optional passphrase for the supplied key
        :param expiration: the expiration time of the cert in ISO 8601 format
        :param name: a friendly name for the cert

        :returns: the container_ref of the stored cert
        :raises Exception: if certificate storage fails
        """
        connection = self.auth.get_barbican_client(context.project_id)

        LOG.info("Storing certificate container '%s' in Barbican.", name)

        certificate_secret = None
        private_key_secret = None
        intermediates_secret = None
        pkp_secret = None

        try:
            certificate_secret = connection.secrets.create(
                payload=certificate,
                expiration=expiration,
                name="Certificate"
            )
            private_key_secret = connection.secrets.create(
                payload=private_key,
                expiration=expiration,
                name="Private Key"
            )
            certificate_container = connection.containers.create_certificate(
                name=name,
                certificate=certificate_secret,
                private_key=private_key_secret
            )
            if intermediates:
                intermediates_secret = connection.secrets.create(
                    payload=intermediates,
                    expiration=expiration,
                    name="Intermediates"
                )
                certificate_container.intermediates = intermediates_secret
            if private_key_passphrase:
                pkp_secret = connection.secrets.create(
                    payload=private_key_passphrase,
                    expiration=expiration,
                    name="Private Key Passphrase"
                )
                certificate_container.private_key_passphrase = pkp_secret

            certificate_container.store()
            return certificate_container.container_ref
        except Exception as e:
            for i in [certificate_secret, private_key_secret,
                      intermediates_secret, pkp_secret]:
                if i and i.secret_ref:
                    old_ref = i.secret_ref
                    try:
                        i.delete()
                        LOG.info('Deleted secret %s (%s) during rollback.',
                                 i.name, old_ref)
                    except Exception:
                        LOG.warning('Failed to delete %s (%s) during '
                                    'rollback. This might not be a problem.',
                                    i.name, old_ref)
            with excutils.save_and_reraise_exception():
                LOG.error('Error storing certificate data: %s', str(e))
        return None

    def get_cert(self, context, cert_ref, resource_ref=None, check_only=False,
                 service_name=None):
        """Retrieves the specified cert and registers as a consumer.

        :param context: Oslo context of the request
        :param cert_ref: the UUID of the cert to retrieve
        :param resource_ref: Full HATEOAS reference to the consuming resource
        :param check_only: Read Certificate data without registering
        :param service_name: Friendly name for the consuming service

        :return: octavia.certificates.common.Cert representation of the
                 certificate data
        :raises Exception: if certificate retrieval fails
        """
        connection = self.auth.get_barbican_client(context.project_id)

        LOG.info('Loading certificate container %s from Barbican.', cert_ref)
        try:
            if check_only:
                cert_container = connection.containers.get(
                    container_ref=cert_ref
                )
            else:
                cert_container = connection.containers.register_consumer(
                    container_ref=cert_ref,
                    name=service_name,
                    url=resource_ref
                )
            barbican_cert = barbican_common.BarbicanCert(cert_container)

            LOG.debug('Validating certificate data for %s.', cert_ref)
            cert_parser.validate_cert(
                barbican_cert.get_certificate(),
                private_key=barbican_cert.get_private_key(),
                private_key_passphrase=(
                    barbican_cert.get_private_key_passphrase()),
                intermediates=barbican_cert.get_intermediates())
            LOG.debug('Certificate data validated for %s.', cert_ref)

            return barbican_cert
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error('Error getting cert %s: %s', cert_ref, str(e))
        return None

    def delete_cert(self, context, cert_ref, resource_ref, service_name=None):
        """Deregister as a consumer for the specified cert.

        :param context: Oslo context of the request
        :param cert_ref: the UUID of the cert to retrieve
        :param resource_ref: Full HATEOAS reference to the consuming resource
        :param service_name: Friendly name for the consuming service

        :raises Exception: if deregistration fails
        """
        connection = self.auth.get_barbican_client(context.project_id)

        LOG.info('Deregistering as a consumer of %s in Barbican.', cert_ref)
        try:
            connection.containers.remove_consumer(
                container_ref=cert_ref,
                name=service_name,
                url=resource_ref
            )
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error('Error deregistering as a consumer of %s: %s',
                          cert_ref, str(e))

    def set_acls(self, context, cert_ref):
        connection = self.auth.get_barbican_client(context.project_id)
        try:
            cert_container = connection.containers.get(
                container_ref=cert_ref
            )
        except Exception:
            # If the containers.get failed, it was probably because it isn't
            # legacy so we will skip this step
            return
        self.auth.ensure_secret_access(
            context, cert_container.certificate.secret_ref)
        self.auth.ensure_secret_access(
            context, cert_container.private_key.secret_ref)
        if cert_container.private_key_passphrase:
            self.auth.ensure_secret_access(
                context,
                cert_container.private_key_passphrase.secret_ref)
        if cert_container.intermediates:
            self.auth.ensure_secret_access(
                context, cert_container.intermediates.secret_ref)

    def unset_acls(self, context, cert_ref):
        connection = self.auth.get_barbican_client(context.project_id)
        try:
            cert_container = connection.containers.get(
                container_ref=cert_ref
            )
        except Exception:
            # If the containers.get failed, it was probably because it isn't
            # legacy so we will skip this step
            return
        self.auth.revoke_secret_access(
            context, cert_container.certificate.secret_ref)
        self.auth.revoke_secret_access(
            context, cert_container.private_key.secret_ref)
        if cert_container.private_key_passphrase:
            self.auth.revoke_secret_access(
                context,
                cert_container.private_key_passphrase.secret_ref)
        if cert_container.intermediates:
            self.auth.revoke_secret_access(
                context, cert_container.intermediates.secret_ref)

    def get_secret(self, context, secret_ref):
        # The legacy driver doesn't need get_secret
        return None



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log as logging

from octavia.common import clients
from octavia.common import constants
from octavia.common import exceptions
from octavia.image import image_base

LOG = logging.getLogger(__name__)

CONF = cfg.CONF


class ImageManager(image_base.ImageBase):
    '''Image implementation of virtual machines via Glance.'''

    def __init__(self):
        super().__init__()
        # Must initialize glance api
        self._glance_client = clients.GlanceAuth.get_glance_client(
            service_name=CONF.glance.service_name,
            endpoint=CONF.glance.endpoint,
            region=CONF.glance.region_name,
            endpoint_type=CONF.glance.endpoint_type,
            insecure=CONF.glance.insecure,
            cacert=CONF.glance.ca_certificates_file
        )
        self.manager = self._glance_client.images

    def get_image_id_by_tag(self, image_tag, image_owner=None):
        """Get image ID by image tag and owner

        :param image_tag: image tag
        :param image_owner: optional image owner
        :raises: ImageGetException if no images found with given tag
        :return: image id
        """
        filters = {'tag': [image_tag],
                   'status': constants.GLANCE_IMAGE_ACTIVE}
        if image_owner:
            filters.update({'owner': image_owner})

        images = list(self.manager.list(
            filters=filters, sort='created_at:desc', limit=2))

        if not images:
            raise exceptions.ImageGetException(tag=image_tag)
        image_id = images[0]['id']
        num_images = len(images)
        if num_images > 1:
            LOG.warning("A single Glance image should be tagged with %(tag)s "
                        "tag, but at least two were found. Using "
                        "%(image_id)s.",
                        {'tag': image_tag, 'image_id': image_id})
        return image_id
